import random
from pathlib import Path

from rlbot.matchconfig.match_config import PlayerConfig
from rlbot.parsing.agent_config_parser import load_bot_appearance, create_looks_configurations

PSYONIX_PRESET_DIRECTORY = Path(__file__).parent / "psyonix_presets"
PSYONIX_PRESET_LIST = list(PSYONIX_PRESET_DIRECTORY.glob("*.cfg"))
random.shuffle(PSYONIX_PRESET_LIST)


def set_random_psyonix_bot_preset(player: PlayerConfig):
    """
    Sets the name and loadout to a randomly selected psyonix bot preset.
    """
    loadout_file = PSYONIX_PRESET_LIST[0]
    PSYONIX_PRESET_LIST.append(PSYONIX_PRESET_LIST.pop(0))  # Move preset to the end of the list.
    loadout_config = create_looks_configurations().parse_file(loadout_file)
    player.loadout_config = load_bot_appearance(loadout_config, player.team)
    player.name = loadout_file.name.split("_")[1].title()
    player.name = ["Rookie ", "Pro ", ""][min(int(abs(player.bot_skill) * 3), 2)] + player.name



from dataclasses import dataclass
from typing import List

from rlbot.utils.structures.game_data_struct import Physics

MAX_SLICES = 360

@dataclass
class Slice:
    physics: Physics
    game_seconds: float


@dataclass
class BallPrediction:
    slices: List[Slice]
    num_slices: int



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class RenderGroup(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRenderGroup(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RenderGroup()
        x.Init(buf, n + offset)
        return x

    # RenderGroup
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RenderGroup
    def RenderMessages(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .RenderMessage import RenderMessage
            obj = RenderMessage()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RenderGroup
    def RenderMessagesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

# /// The id of the render group
    # RenderGroup
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def RenderGroupStart(builder): builder.StartObject(2)
def RenderGroupAddRenderMessages(builder, renderMessages): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(renderMessages), 0)
def RenderGroupStartRenderMessagesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RenderGroupAddId(builder, id): builder.PrependInt32Slot(1, id, 0)
def RenderGroupEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DropShotBallInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDropShotBallInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DropShotBallInfo()
        x.Init(buf, n + offset)
        return x

    # DropShotBallInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DropShotBallInfo
    def AbsorbedForce(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # DropShotBallInfo
    def DamageIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # DropShotBallInfo
    def ForceAccumRecent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def DropShotBallInfoStart(builder): builder.StartObject(3)
def DropShotBallInfoAddAbsorbedForce(builder, absorbedForce): builder.PrependFloat32Slot(0, absorbedForce, 0.0)
def DropShotBallInfoAddDamageIndex(builder, damageIndex): builder.PrependInt32Slot(1, damageIndex, 0)
def DropShotBallInfoAddForceAccumRecent(builder, forceAccumRecent): builder.PrependFloat32Slot(2, forceAccumRecent, 0.0)
def DropShotBallInfoEnd(builder): return builder.EndObject()



from typing import Optional
from urllib.parse import ParseResult as URL

from flatbuffers import Builder
from rlbot.agents.rlbot_runnable import RLBotRunnable, LOCATIONS_HEADER, DETAILS_HEADER
from rlbot.botmanager.helper_process_request import HelperProcessRequest
from rlbot.matchcomms.client import MatchcommsClient
from rlbot.messages.flat import MatchSettings, ControllerState, PlayerInput
from rlbot.parsing.custom_config import ConfigObject
from rlbot.utils.game_state_util import GameState
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.rendering.rendering_manager import RenderingManager
from rlbot.utils.structures.ball_prediction_struct import BallPrediction
from rlbot.utils.structures.game_data_struct import GameTickPacket, FieldInfoPacket
from rlbot.utils.structures.legacy_data_v3 import convert_to_legacy_v3
from rlbot.utils.structures.quick_chats import QuickChats
from rlbot.utils.structures.rigid_body_struct import RigidBodyTick

BOT_CONFIG_MODULE_HEADER = LOCATIONS_HEADER
BOT_CONFIG_AGENT_HEADER = 'Bot Parameters'
BOT_CONFIG_DETAILS_HEADER = DETAILS_HEADER
PYTHON_FILE_KEY = 'python_file'
SUPPORTS_STANDALONE = 'supports_standalone'
LOADOUT_GENERATOR_FILE_KEY = 'loadout_generator'
LOGO_FILE_KEY = 'logo_file'
LOOKS_CONFIG_KEY = 'looks_config'
BOT_NAME_KEY = "name"
SUPPORTS_EARLY_START_KEY = "supports_early_start"
MAXIMUM_TICK_RATE_PREFERENCE_KEY = "maximum_tick_rate_preference"


class SimpleControllerState:
    """
    Building flatbuffer objects is verbose and error prone. This class provides a friendlier
    interface to bot makers.
    """

    def __init__(self,
                 steer: float = 0.0,
                 throttle: float = 0.0,
                 pitch: float = 0.0,
                 yaw: float = 0.0,
                 roll: float = 0.0,
                 jump: bool = False,
                 boost: bool = False,
                 handbrake: bool = False,
                 use_item: bool = False):
        """
        :param steer:    Range: -1 .. 1, negative=left, positive=right
        :param throttle: Range: -1 .. 1, negative=backward, positive=forward
        :param pitch:    Range: -1 .. 1, negative=nose-down, positive=nose-up
        :param yaw:      Range: -1 .. 1, negative=nose-left, positive=nose-right
        :param roll:     Range: -1 .. 1, negative=anticlockwise, positive=clockwise  (when looking forwards along the car)
        :param jump: Analogous to the jump button in game.
        :param boost: Analogous to the boost button in game.
        :param handbrake: Analogous to the handbrake button in game.
        :param use_item: Analogous to the use item button (from rumble) in game.
        """
        self.steer = steer
        self.throttle = throttle
        self.pitch = pitch
        self.yaw = yaw
        self.roll = roll
        self.jump = jump
        self.boost = boost
        self.handbrake = handbrake
        self.use_item = use_item

    def to_flatbuffer(self, player_index: int) -> Builder:
        builder = Builder(100)
        ControllerState.ControllerStateStart(builder)
        ControllerState.ControllerStateAddSteer(builder, self.steer)
        ControllerState.ControllerStateAddThrottle(builder, self.throttle)
        ControllerState.ControllerStateAddPitch(builder, self.pitch)
        ControllerState.ControllerStateAddYaw(builder, self.yaw)
        ControllerState.ControllerStateAddRoll(builder, self.roll)
        ControllerState.ControllerStateAddJump(builder, self.jump)
        ControllerState.ControllerStateAddBoost(builder, self.boost)
        ControllerState.ControllerStateAddHandbrake(builder, self.handbrake)
        cs_offset = ControllerState.ControllerStateEnd(builder)
        PlayerInput.PlayerInputStart(builder)
        PlayerInput.PlayerInputAddPlayerIndex(builder, player_index)
        PlayerInput.PlayerInputAddControllerState(builder, cs_offset)
        pi_offset = PlayerInput.PlayerInputEnd(builder)
        builder.Finish(pi_offset)
        return builder


class BaseAgent(RLBotRunnable):
    # 'team' is an integer: 0 towards positive goal, 1 towards negative goal.
    # 0 is blue team, 1 is orange team
    team = None

    # 'index' is an integer: it is index at which the bot appears inside game_tick_packet.game_cars
    index = None

    # passed in by the bot manager
    __quick_chat_func = None
    __field_info_func = None
    __game_state_func = None
    __get_rigid_body_tick_func = None
    __match_settings_func = None
    renderer: RenderingManager = None
    matchcomms_root: URL = None

    def __init__(self, name, team, index):
        super().__init__(name)
        self.team = team
        self.index = index
        self.logger = get_logger(f'bot{index}')
        self.spawn_id = -1

    def get_output(self, game_tick_packet: GameTickPacket) -> SimpleControllerState:
        """
        Where all the logic of your bot gets its input and returns its output.
        :param game_tick_packet: see https://github.com/drssoccer55/RLBot/wiki/Input-and-Output-Data-(current)
        :return: [throttle, steer, pitch, yaw, roll, jump, boost, handbrake]
        """
        return SimpleControllerState()

    def send_quick_chat(self, team_only, quick_chat):
        """
        Sends a quick chat to the other bots.
        If it is QuickChats.CHAT_NONE or None it does not send a quick chat to other bots.
        :param team_only: either True or False, this says if the quick chat should only go to team members.
        :param quick_chat: The quick chat selection, available chats are defined in quick_chats.py
        """
        if quick_chat == QuickChats.CHAT_NONE or quick_chat is None:
            return
        self.__quick_chat_func(team_only, quick_chat)

    def handle_quick_chat(self, index, team, quick_chat):
        """
        Handles a quick chat from another bot.
        This will not receive any chats that this bot sends out.
        Currently does nothing, override to add functionality.
        :param index: Returns the index in the list of game cars that sent the quick chat
        :param team: Which team this player is on
        :param quick_chat: What the quick chat selection was
        """
        pass

    def get_field_info(self):
        """Gets the information about the field.
        This does not change during a match so it only needs to be called once after the everything is loaded."""
        return self.__field_info_func()

    def get_rigid_body_tick(self) -> RigidBodyTick:
        """Get the most recent state of the physics engine."""
        return self.__get_rigid_body_tick_func()

    def set_game_state(self, game_state: GameState):
        """CHEAT: Change the rocket league game to the given game_state"""
        self.__game_state_func(game_state)

    def get_ball_prediction(self):
        """DEPRECATED! Please use get_ball_prediction_struct instead, because this is going away soon!"""
        return self.__ball_prediction_func()

    def get_ball_prediction_struct(self) -> BallPrediction:
        """Fetches a prediction of where the ball will go during the next few seconds."""
        return self.__ball_prediction_struct_func()

    def get_match_settings(self) -> MatchSettings:
        """Gets the current match settings in flatbuffer format. Useful for determining map, game mode,
        mutator settings, etc."""
        return self.__match_settings_func()

    _matchcomms: Optional[MatchcommsClient] = None
    @property
    def matchcomms(self) -> MatchcommsClient:
        """
        Gets a client to send and recieve messages to other participants in the match (e.g. bots, trainer)
        """
        if self.matchcomms_root is None:
            # We had a choice here between raising the exception (fail fast) or printing a message and
            # returning None since missing matchcomms might be harmless / unused by the bot. Failing fast
            # is almost always the best policy, AND in the one case we've seen so far, returning None would
            # not have saved the situation - https://github.com/VirxEC/VirxERLU/blob/master/match_comms.py
            raise ValueError("Your bot tried to access matchcomms but matchcomms_root is None! This "
                             "may be due to manually running a bot in standalone mode without passing the "
                             "--matchcomms-url argument. That's a fine thing to do, and if it's safe to "
                             "ignore matchcomms in your case then go ahead and wrap your matchcomms access "
                             "in a try-except, or do a check first for whether matchcomms_root is None.")
        if self._matchcomms is None:
            self._matchcomms = MatchcommsClient(self.matchcomms_root)
        return self._matchcomms  # note: _matchcomms.close() is called by the bot_manager.

    def load_config(self, config_object_header):
        """
        Loads a config object this is called after the constructor but before anything else inside the bot.
        :param config_object: This is a config object that has headers, and values for custom agent configuration.
        """
        pass

    def get_extra_pids(self):
        """
        Gets the list of process ids that should be marked as high priority.
        :return: A list of process ids that are used by this bot in addition to the ones inside the python process.
        """
        return []

    def get_helper_process_request(self) -> HelperProcessRequest:
        """
        If your bot needs a helper process which can be shared, e.g. with other bots of the same type,
        you may override this to return a HelperProcessRequest.
        """
        return None

    def convert_output_to_v4(self, controller_input):
        """Converts a v3 output to a v4 controller state"""
        player_input = SimpleControllerState()
        player_input.throttle = controller_input[0]
        player_input.steer = controller_input[1]
        player_input.pitch = controller_input[2]
        player_input.yaw = controller_input[3]
        player_input.roll = controller_input[4]
        player_input.jump = controller_input[5]
        player_input.boost = controller_input[6]
        player_input.handbrake = controller_input[7]

        return player_input

    def convert_packet_to_v3(self, game_tick_packet: GameTickPacket, field_info_packet: FieldInfoPacket = None):
        """Converts the current game tick packet to v3
        :param game_tick_packet a game tick packet in the v4 struct format.
        :param field_info_packet a field info packet in the v4 struct format. Optional. If this is not supplied,
        none of the boost locations will be filled in.
        :return: A v3 version of the game tick packet"""
        return convert_to_legacy_v3(game_tick_packet, field_info_packet)

    def is_hot_reload_enabled(self):
        """
        If true, the framework will watch all your python files for modifications and instantly reload your bot
        so that the logic changes take effect. You may wish to disable this if you're concerned about performance.
        """
        return True

    ############
    #  Methods that should not be called or changed by subclasses
    ############

    def _register_quick_chat(self, quick_chat_func):
        """
        Registers the send quick chat function.
        This should not be overwritten by the agent.
        """
        self.__quick_chat_func = quick_chat_func

    def _register_field_info(self, field_info_func):
        """
        Sets the function to grab field information from the interface.
        This should not be overwritten by the agent.
        """
        self.__field_info_func = field_info_func

    def _register_get_rigid_body_tick(self, get_rigid_body_tick_func):
        self.__get_rigid_body_tick_func = get_rigid_body_tick_func

    def _register_set_game_state(self, game_state_func):
        self.__game_state_func = game_state_func

    def _register_ball_prediction(self, ball_prediction_func):
        """
        Deprecated.  __ball_prediction_struct_func will be used instead.

        Sets the function to grab ball predictions from the interface.
        This should not be overwritten by the agent.
        """
        self.__ball_prediction_func = ball_prediction_func

    def _register_ball_prediction_struct(self, ball_prediction_func):
        """
        Sets the function to grab ball predictions from the interface.
        This should not be overwritten by the agent.
        """
        self.__ball_prediction_struct_func = ball_prediction_func

    def _register_match_settings_func(self, match_settings_func):
        """
        Sets the function to grab match settings from the interface.
        This should not be overwritten by the agent.
        """
        self.__match_settings_func = match_settings_func

    def _set_renderer(self, renderer: RenderingManager):
        self.renderer = renderer

    def _set_spawn_id(self, spawn_id: int):
        self.spawn_id = spawn_id

    # Information about @classmethod: https://docs.python.org/3/library/functions.html#classmethod
    @classmethod
    def base_create_agent_configurations(cls) -> ConfigObject:
        """
        This is used when initializing agent config via builder pattern.
        It also calls `create_agent_configurations` that can be used by BaseAgent subclasses for custom configs.
        :return: Returns an instance of a ConfigObject object.
        """

        config = super().base_create_agent_configurations()
        location_config = config.get_header(LOCATIONS_HEADER)

        location_config.add_value(PYTHON_FILE_KEY, str,
                                  description="Bot's python file.\nOnly need this if RLBot controlled")
        location_config.add_value(LOOKS_CONFIG_KEY, str,
                                  description='Path to loadout config from runner')
        location_config.add_value(LOADOUT_GENERATOR_FILE_KEY, str,
                                  description="A file that provide dynamic bot loadouts (optional).")
        location_config.add_value(MAXIMUM_TICK_RATE_PREFERENCE_KEY, int, default=60,
                                  description="The maximum number of ticks per second that your bot wishes to receive.")
        location_config.add_value(SUPPORTS_STANDALONE, bool,
                                  description="True if the bot can accept args and run as a standalone process.")

        cls.create_agent_configurations(config)

        return config



import ctypes

from rlbot.utils.structures.game_data_struct import Physics

MAX_SLICES = 360


class Slice(ctypes.Structure):
    _fields_ = [("physics", Physics),
                ("game_seconds", ctypes.c_float)]


class BallPrediction(ctypes.Structure):
    _fields_ = [("slices", Slice * MAX_SLICES),
                ('num_slices', ctypes.c_int)]



import asyncio
from asyncio import StreamReader, StreamWriter

from flatbuffers.builder import Builder
from rlbot.socket.socket_manager import SocketRelay, int_to_bytes, SocketDataType, SocketMessage, int_from_bytes


class SocketRelayAsyncio(SocketRelay):
    """
    A version of SocketRelay that uses asyncio. This is useful if you also want to run a websocket or something.

    Usage example:

    socket_relay = SocketRelayAsyncio()
    socket_relay.packet_handlers.append(self.whatever)

    relay_run = socket_relay.connect_and_run(wants_quick_chat=True, wants_game_messages=True,
                                                     wants_ball_predictions=True)
    websocket_server = websockets.serve(self.handle_connection, "localhost", 8765)
    asyncio.get_event_loop().run_until_complete(websocket_server)
    asyncio.get_event_loop().run_until_complete(relay_run)
    asyncio.get_event_loop().run_forever()
    """
    def __init__(self):
        super().__init__()
        self.writer: StreamWriter = None
        self.reader: StreamReader = None

    def send_flatbuffer(self, builder: Builder, data_type: SocketDataType):
        flatbuffer_bytes = builder.Output()
        size = len(flatbuffer_bytes)
        message = int_to_bytes(data_type) + int_to_bytes(size) + flatbuffer_bytes
        self.writer.write(message)

    async def read_from_socket_async(self) -> SocketMessage:
        type_int = int_from_bytes(await self.reader.readexactly(2))
        data_type = SocketDataType(type_int)
        size = int_from_bytes(await self.reader.readexactly(2))
        data = await self.reader.readexactly(size)
        return SocketMessage(data_type, data)

    async def wait_and_handle(self):
        incoming_message = await self.read_from_socket_async()
        self.handle_incoming_message(incoming_message)
        if self._should_continue:
            asyncio.create_task(self.wait_and_handle())
        else:
            await self.writer.drain()
            self.writer.close()

    async def connect_and_run(self, wants_quick_chat, wants_game_messages, wants_ball_predictions):
        """
        Connects to the socket and begins a loop that reads messages and calls any handlers
        that have been registered. Connect and run are combined into a single method because
        currently bad things happen if the buffer is allowed to fill up.
        """

        self.reader, self.writer, = await asyncio.open_connection('127.0.0.1', 23234)
        self.is_connected = True
        self.logger.info("Connected!")
        for handler in self.on_connect_handlers:
            handler()

        builder = self.make_ready_message(wants_ball_predictions, wants_game_messages, wants_quick_chat)
        self.send_flatbuffer(builder, SocketDataType.READY_MESSAGE)

        await self.wait_and_handle()



import sys

# This module exists to give users an indication that they need to have
# a version of python compatible with the RLBot framework.
# Otherwise people might find out by in-the-guts error messages
# after quite a while of the runner launching.
minimum_python_version = (3, 6)

# Deliberately using old string formatting for compatibility.
error_string = """You appear to be using an old version of Python: %s
 RLBot requires Python %d.%d or later.
 After installing, ensure your environment point to the new Python version, then run setup.bat""" % (
    (sys.version,) + minimum_python_version)


def check_python_version():
    assert sys.version_info >= minimum_python_version, error_string



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// The state of a rigid body in Rocket League's physics engine.
# /// This gets updated in time with the physics tick, not the rendering framerate.
# /// The frame field will be incremented every time the physics engine ticks.
class RigidBodyState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRigidBodyState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RigidBodyState()
        x.Init(buf, n + offset)
        return x

    # RigidBodyState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RigidBodyState
    def Frame(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # RigidBodyState
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RigidBodyState
    def Rotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Quaternion import Quaternion
            obj = Quaternion()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RigidBodyState
    def Velocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RigidBodyState
    def AngularVelocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def RigidBodyStateStart(builder): builder.StartObject(5)
def RigidBodyStateAddFrame(builder, frame): builder.PrependInt32Slot(0, frame, 0)
def RigidBodyStateAddLocation(builder, location): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def RigidBodyStateAddRotation(builder, rotation): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(rotation), 0)
def RigidBodyStateAddVelocity(builder, velocity): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(velocity), 0)
def RigidBodyStateAddAngularVelocity(builder, angularVelocity): builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(angularVelocity), 0)
def RigidBodyStateEnd(builder): return builder.EndObject()



"""
External code is encouraged to use these functions directly. It is part of our public API.

Avoid moving this file or changing any of the function signatures!
"""
from rlbot.utils.file_util import contains_locked_file, get_rlbot_directory
try:
    import httplib
except:
    import http.client as httplib


def is_safe_to_upgrade():
    """
    Reports whether it's currently safe to upgrade the rlbot package, e.g. via pip. An example of when it's
    unsafe is when one of the dlls is in use, which can cause the upgrade to fail half way through.
    See https://github.com/RLBot/RLBot/issues/130
    """
    return not contains_locked_file(get_rlbot_directory())

# https://stackoverflow.com/questions/3764291/checking-network-connection


def have_internet():
    conn = httplib.HTTPConnection("www.google.com", timeout=5)
    try:
        conn.request("HEAD", "/")
        return True
    except:
        return False
    finally:
        conn.close()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class Float(object):
    __slots__ = ['_tab']

    # Float
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Float
    def Val(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))

def CreateFloat(builder, val):
    builder.Prep(4, 4)
    builder.PrependFloat32(val)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DesiredPhysics(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDesiredPhysics(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DesiredPhysics()
        x.Init(buf, n + offset)
        return x

    # DesiredPhysics
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DesiredPhysics
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Vector3Partial import Vector3Partial
            obj = Vector3Partial()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredPhysics
    def Rotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .RotatorPartial import RotatorPartial
            obj = RotatorPartial()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredPhysics
    def Velocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Vector3Partial import Vector3Partial
            obj = Vector3Partial()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredPhysics
    def AngularVelocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Vector3Partial import Vector3Partial
            obj = Vector3Partial()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def DesiredPhysicsStart(builder): builder.StartObject(4)
def DesiredPhysicsAddLocation(builder, location): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def DesiredPhysicsAddRotation(builder, rotation): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(rotation), 0)
def DesiredPhysicsAddVelocity(builder, velocity): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(velocity), 0)
def DesiredPhysicsAddAngularVelocity(builder, angularVelocity): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(angularVelocity), 0)
def DesiredPhysicsEnd(builder): return builder.EndObject()



import ctypes
import math

from rlbot.utils.structures import game_data_struct
from rlbot.utils.structures.game_data_struct import MAX_BOOSTS
from rlbot.utils.structures.start_match_structures import MAX_NAME_LENGTH, MAX_PLAYERS

UNREAL_ROT_PER_RADIAN = 32768 / math.pi


class Vector3(ctypes.Structure):
    _fields_ = [("X", ctypes.c_float),
                ("Y", ctypes.c_float),
                ("Z", ctypes.c_float)]


class Rotator(ctypes.Structure):
    _fields_ = [("Pitch", ctypes.c_int),
                ("Yaw", ctypes.c_int),
                ("Roll", ctypes.c_int)]


class Touch(ctypes.Structure):
    _fields_ = [("wPlayerName", ctypes.c_wchar * MAX_NAME_LENGTH),
                ("fTimeSeconds", ctypes.c_float),
                ("sHitLocation", Vector3),
                ("sHitNormal", Vector3)]


class ScoreInfo(ctypes.Structure):
    _fields_ = [("Score", ctypes.c_int),
                ("Goals", ctypes.c_int),
                ("OwnGoals", ctypes.c_int),
                ("Assists", ctypes.c_int),
                ("Saves", ctypes.c_int),
                ("Shots", ctypes.c_int),
                ("Demolitions", ctypes.c_int)]


class PlayerInfo(ctypes.Structure):
    _fields_ = [("Location", Vector3),
                ("Rotation", Rotator),
                ("Velocity", Vector3),
                ("AngularVelocity", Vector3),
                ("Score", ScoreInfo),
                ("bDemolished", ctypes.c_bool),
                # True if your wheels are on the ground, the wall, or the ceiling. False if you're midair or turtling.
                ("bOnGround", ctypes.c_bool),
                ("bSuperSonic", ctypes.c_bool),
                ("bBot", ctypes.c_bool),
                # True if the player has jumped. Falling off the ceiling / driving off the goal post does not count.
                ("bJumped", ctypes.c_bool),
                # True if player has double jumped. False does not mean you have a jump remaining, because the
                # aerial timer can run out, and that doesn't affect this flag.
                ("bDoubleJumped", ctypes.c_bool),
                ("wName", ctypes.c_wchar * MAX_NAME_LENGTH),
                ("Team", ctypes.c_ubyte),
                ("Boost", ctypes.c_int)]


class BallInfo(ctypes.Structure):
    _fields_ = [("Location", Vector3),
                ("Rotation", Rotator),
                ("Velocity", Vector3),
                ("AngularVelocity", Vector3),
                ("Acceleration", Vector3),
                ("LatestTouch", Touch)]


class BoostInfo(ctypes.Structure):
    _fields_ = [("Location", Vector3),
                ("bActive", ctypes.c_bool),
                ("Timer", ctypes.c_float)]


class GameInfo(ctypes.Structure):
    _fields_ = [("TimeSeconds", ctypes.c_float),
                ("GameTimeRemaining", ctypes.c_float),
                ("bOverTime", ctypes.c_bool),
                ("bUnlimitedTime", ctypes.c_bool),
                # True when cars are allowed to move, and during the pause menu. False during replays.
                ("bRoundActive", ctypes.c_bool),
                # Only false during a kickoff, when the car is allowed to move, and the ball has not been hit,
                # and the game clock has not started yet. If both players sit still, game clock will eventually
                # start and this will become true.
                ("bBallHasBeenHit", ctypes.c_bool),
                # Turns true after final replay, the moment the 'winner' screen appears. Remains true during next match
                # countdown. Turns false again the moment the 'choose team' screen appears.
                ("bMatchEnded", ctypes.c_bool)]


# On the c++ side this struct has a long at the beginning for locking.
# This flag is removed from this struct so it isn't visible to users.
class GameTickPacket(ctypes.Structure):
    _fields_ = [("gamecars", PlayerInfo * MAX_PLAYERS),
                ("numCars", ctypes.c_int),
                ("gameBoosts", BoostInfo * MAX_BOOSTS),
                ("numBoosts", ctypes.c_int),
                ("gameball", BallInfo),
                ("gameInfo", GameInfo)]


def convert_to_legacy_v3(
        game_tick_packet: game_data_struct.GameTickPacket,
        field_info_packet: game_data_struct.FieldInfoPacket = None):
    """
    Returns a legacy packet from v3
    :param game_tick_packet a game tick packet in the v4 struct format.
    :param field_info_packet a field info packet in the v4 struct format. Optional. If this is not supplied,
    none of the boost locations will be filled in.
    """
    legacy_packet = GameTickPacket()

    legacy_packet.numBoosts = game_tick_packet.num_boost
    legacy_packet.numCars = game_tick_packet.num_cars

    for i in range(game_tick_packet.num_cars):
        convert_player_info(legacy_packet.gamecars[i], game_tick_packet.game_cars[i])

    for i in range(game_tick_packet.num_boost):
        convert_boost_info(legacy_packet.gameBoosts[i], game_tick_packet.game_boosts[i])
        if field_info_packet is not None:
            convert_vector(legacy_packet.gameBoosts[i].Location, field_info_packet.boost_pads[i].location)

    convert_ball_info(legacy_packet.gameball, game_tick_packet.game_ball)

    convert_game_info(legacy_packet.gameInfo, game_tick_packet.game_info)

    return legacy_packet


def convert_game_info(legacy_game_info, game_info):
    legacy_game_info.TimeSeconds = game_info.seconds_elapsed
    legacy_game_info.GameTimeRemaining = game_info.game_time_remaining
    legacy_game_info.bOverTime = game_info.is_overtime
    legacy_game_info.bUnlimitedTime = game_info.is_unlimited_time
    legacy_game_info.bRoundActive = game_info.is_round_active
    legacy_game_info.bBallHasBeenHit = not game_info.is_kickoff_pause
    legacy_game_info.bMatchEnded = game_info.is_match_ended


def convert_player_info(legacy_player_info, player_info):
    convert_vector(legacy_player_info.Location, player_info.physics.location)
    convert_rotator(legacy_player_info.Rotation, player_info.physics.rotation)
    convert_vector(legacy_player_info.Velocity, player_info.physics.velocity)
    convert_vector(legacy_player_info.AngularVelocity, player_info.physics.angular_velocity)

    # score
    legacy_player_info.Score.Score = player_info.score_info.score
    legacy_player_info.Score.Goals = player_info.score_info.goals
    legacy_player_info.Score.Saves = player_info.score_info.saves
    legacy_player_info.Score.Shots = player_info.score_info.shots
    legacy_player_info.Score.Assists = player_info.score_info.assists
    legacy_player_info.Score.OwnGoals = player_info.score_info.own_goals
    legacy_player_info.Score.Demolitions = player_info.score_info.demolitions

    legacy_player_info.bDemolished = player_info.is_demolished
    legacy_player_info.bOnGround = player_info.has_wheel_contact
    legacy_player_info.bSuperSonic = player_info.is_super_sonic
    legacy_player_info.bBot = player_info.is_bot
    legacy_player_info.bJumped = player_info.jumped
    legacy_player_info.bDoubleJumped = player_info.double_jumped
    legacy_player_info.wName = player_info.name
    legacy_player_info.Team = player_info.team
    legacy_player_info.Boost = player_info.boost


def convert_ball_info(legacy_ball_info, ball_info):
    convert_vector(legacy_ball_info.Location, ball_info.physics.location)
    convert_rotator(legacy_ball_info.Rotation, ball_info.physics.rotation)
    convert_vector(legacy_ball_info.Velocity, ball_info.physics.velocity)
    convert_vector(legacy_ball_info.AngularVelocity, ball_info.physics.angular_velocity)
    legacy_ball_info.LatestTouch.wPlayerName = ball_info.latest_touch.player_name
    legacy_ball_info.LatestTouch.fTimeSeconds = ball_info.latest_touch.time_seconds
    convert_vector(legacy_ball_info.LatestTouch.sHitLocation, ball_info.latest_touch.hit_location)
    convert_vector(legacy_ball_info.LatestTouch.sHitNormal, ball_info.latest_touch.hit_normal)


def convert_boost_info(legacy_boost_info, boost_info):
    legacy_boost_info.bActive = boost_info.is_active
    legacy_boost_info.Timer = boost_info.timer


def convert_vector(legacy_vector, vector):
    legacy_vector.X = vector.x
    legacy_vector.Y = vector.y
    legacy_vector.Z = vector.z


def convert_rotator(legacy_rotator, rotator):
    legacy_rotator.Pitch = int(rotator.pitch * UNREAL_ROT_PER_RADIAN)
    legacy_rotator.Yaw = int(rotator.yaw * UNREAL_ROT_PER_RADIAN)
    legacy_rotator.Roll = int(rotator.roll * UNREAL_ROT_PER_RADIAN)



from configparser import RawConfigParser
from pathlib import Path
from typing import Union
import re

import os


class ConfigObject:
    """
    Represents a parsed config file.
    Could be either of rlbot match or appearance config.
    The config schema is defined by calls to add_header().
    """

    def __init__(self):
        self.headers = {}
        self.raw_config_parser = None  # set by calling parse_file
        self.config_directory = None

    def __getitem__(self, x):
        return self.get_header(x)

    def add_header(self, header_name, header):
        """Adds the given header into this object with the name `header_name`"""
        self.headers[header_name] = header
        return header

    def add_header_name(self, header_name, is_indexed=False):
        """
        Adds a new header with the name header_name
        :param header_name: The name of the header as it would appear in the config
        :param is_indexed: Indicates that the header may contain special 'indexed' attributes, where a series of
        keys that differ only by a numeric suffix are used to represent an array. Example:
        thing_1 = x
        thing_2 = y
        :return: The newly created header.
        """
        header = ConfigHeader()
        header.is_indexed = is_indexed
        self.headers[header_name] = header
        return header

    def set_value(self, header_name, option, value, index=None):
        self.get_header(header_name).set_value(option, value, index)

    def get_header(self, header_name):
        """
        Returns a header with that name, creates it if it does not exist.
        """
        if header_name in self.headers:
            return self.headers[header_name]
        return self.add_header_name(header_name)

    def get(self, section, option, index=None):
        return self.get_header(section).get(option, index=index)

    def getpath(self, section, option, index=None):
        relative_path = self.get_header(section).get(option, index=index)
        if relative_path is None:
            return None
        return os.path.realpath(os.path.join(self.config_directory, relative_path))

    def getint(self, section, option, index=None):
        return self.get_header(section).getint(option, index=index)

    def getboolean(self, section, option, index=None):
        return self.get_header(section).getboolean(option, index=index)

    def getfloat(self, section, option, index=None):
        return self.get_header(section).getfloat(option, index=index)

    def init_indices(self, max_index):
        for header in self.headers.values():
            header.init_indices(max_index)

    def parse_file(self, config: Union[Path, str, RawConfigParser, 'ConfigObject'], max_index=None, config_directory=None):
        """
        Parses the file internally setting values
        :param config: an instance of RawConfigParser or a string to a .cfg file
        :param config_directory: The location of this config file on disk
        :return: None
        """
        self.config_directory = config_directory

        if isinstance(config, Path):
            config = str(config)

        if isinstance(config, str):
            if not os.path.isfile(config):
                raise FileNotFoundError(config)
            self.raw_config_parser = RawConfigParser()
            self.raw_config_parser.read(config, encoding='utf8')
            config = self.raw_config_parser
        elif isinstance(config, RawConfigParser):
            self.raw_config_parser = config
        elif not isinstance(config, ConfigObject):
            raise TypeError("The config should be a String, RawConfigParser of a ConfigObject")
        for header_name, header in self.headers.items():
            try:
                header.parse_file(config[header_name], max_index=max_index, config_directory=config_directory)
            except KeyError:
                pass  # skip this header as it does not exist
        return self

    def reset(self):
        for header_name in self.headers:
            header = self.headers[header_name]
            header.reset()

    def __str__(self):
        string = ''
        for header_name, header in self.headers.items():
            string += '[' + header_name + ']\n' + str(header) + '\n'
        return string

    def copy(self):
        new_object = ConfigObject()
        new_object.config_directory = self.config_directory
        for header_name, header in self.headers.items():
            new_object.add_header(header_name, header.copy())
        return new_object

    def has_section(self, header_name):
        """Returns true if the header exists"""
        return header_name in self.headers

    def get_raw_file(self):
        """Returns the raw file from the parser so it can be used to be parsed by other config objects."""
        return self.raw_config_parser


class ConfigHeader:

    def __init__(self):
        self.values = {}

        # Indicates that the header may contain special 'indexed' attributes, where a series of
        # keys that differ only by a numeric suffix are used to represent an array.
        self.is_indexed = False
        self.max_index = -1
        self.config_directory = None

    def __getitem__(self, x):
        return self.values[x]

    def add_value(self, name, value_type, default=None, description=None, value=None):
        """
        Adds a new value to this config header
        :param name: The name of the value as it would appear in a config file
        :param value_type:  The type of value: bool, str, int, float
        :param default: The value used when the config does not set any value.
        :param description: The human readable description of the value
        :param value: An optional value, if this header is indexed then the value needs to be a list.
        :return: an instance of itself so that you can chain adding values together.
        """
        if description is None:
            description = name
        if value is not None and self.is_indexed and not isinstance(value, list):
            raise Exception('Indexed values must be a list')
        self.values[name] = ConfigValue(value_type, default=default, description=description, value=value)
        return self

    def add_config_value(self, name, value):
        self.values[name] = value

    def set_value(self, option, value, index=None):
        """
        Sets the value on the given option.
        :param option: The name of the option as it appears in the config file
        :param value: The value that is being applied. If this section is indexed then the
        value must be a list (to be applied directly) or you must supply the index parameter,
        which will cause the value to be inserted into an existing list.
        :param index: If the attribute is indexed, we will use this index to insert
        the value you have supplied.
        :return: an instance of itself so that you can chain setting values together.
        """
        if self.is_indexed and index is None and not isinstance(value, list):
            raise TypeError("Value should be a list when not giving an index in an indexed header")
        self.values[option].set_value(value=value, index=index)
        return self

    def get(self, option, index=None):
        return self.values[option].get_value(index=index)

    def getpath(self, option, index=None):
        relative_path = self.values[option].get_value(index=index)
        if relative_path is None:
            return None
        return os.path.realpath(os.path.join(self.config_directory, relative_path))

    def getint(self, option, index=None):
        return int(self.values[option].get_value(index=index))

    def getboolean(self, option, index=None):
        return bool(self.values[option].get_value(index=index))

    def getfloat(self, option, index=None):
        return float(self.values[option].get_value(index=index))

    def init_indices(self, max_index):
        if not self.is_indexed:
            return
        self.max_index = max_index
        for value_name in self.values:
            self.values[value_name].init_indices(max_index)

    def parse_file(self, config_parser, max_index=None, config_directory=None):

        self.config_directory = config_directory

        if self.is_indexed and max_index is None:
            return  # if we do not know the index lets skip instead of crashing

        if not self.is_indexed:
            max_index = None
        else:
            self.max_index = max_index

        for value_name in self.values:
            self.values[value_name].parse_file(config_parser, value_name, max_index=max_index)

    def reset(self):
        for value_name in self.values:
            self.values[value_name].reset()

    def __str__(self):
        string = ''
        for value_name in self.values:
            if self.is_indexed:
                string += self.get_indexed_string(value_name)
                string += '\n'
            else:
                string += self.get_string(value_name)
        return string

    def copy(self):
        new_header = ConfigHeader()
        new_header.is_indexed = self.is_indexed
        new_header.max_index = self.max_index
        for value_name, value in self.values.items():
            new_header.values[value_name] = value.copy()
        new_header.config_directory = self.config_directory
        return new_header

    def get_indexed_string(self, value_name):
        value = self.values[value_name]
        string = value.comment_description() + '\n'
        for i in range(self.max_index):
            string += value_name + '_' + str(i) + ' = ' + str(value.get_value(index=i)) + '\n'
        return string

    def get_string(self, value_name):
        value = self.values[value_name]
        string = value.comment_description() + '\n'
        multiline_safe_value = str(value.get_value()).replace('\n', '\n\t')
        string += value_name + ' = ' + multiline_safe_value + '\n'
        return string


class ConfigValue:
    def __init__(self, value_type, default=None, description="", value=None):
        self.type = value_type
        self.value = value
        self.default = default
        self.description = description

    def get_value(self, index=None):
        """
        Returns the default if value is none.
        :param index:
        :return: A value.
        """

        if self.value is None:
            return self.default
        if index is not None:
            if self.value[index] is None:
                return self.default
            else:
                value = self.value[index]
        else:
            value = self.value

        return value

    def comment_description(self):
        return '# ' + re.sub(r'\n\s*', '\n# ', str(self.description))

    def __str__(self):
        return str(self.get_value()) + '  ' + self.comment_description()

    def copy(self):
        return ConfigValue(self.type, self.default, self.description, self.value)

    def init_indices(self, max_index):
        self.value = [None] * max_index

    def parse_file(self, config_parser, value_name, max_index=None):
        if isinstance(config_parser, ConfigHeader):
            self.value = config_parser[value_name].value
        if max_index is None:
            value = self.get_parser_value(config_parser, value_name)
            self.value = value
        else:
            self.value = []
            for i in range(max_index):
                self.value.append(self.get_parser_value(config_parser, value_name + '_' + str(i)))

    def get_parser_value(self, config_parser, value_name):
        if self.type == bool:
            return config_parser.getboolean(value_name)
        if self.type == int:
            return config_parser.getint(value_name)
        if self.type == float:
            return config_parser.getfloat(value_name)
        return config_parser.get(value_name)

    def set_value(self, value, index=None):
        if index is not None:
            self.value[index] = value
        else:
            self.value = value

    def reset(self):
        self.value = None



from rlbot.agents.base_independent_agent import BaseIndependentAgent


class BaseJavaAgent(BaseIndependentAgent):

    def __init__(self, name, team, index):
        super().__init__(name, team, index)
        raise NotImplementedError(
            f"Cannot run {name} because BaseJavaAgent is deprecated! "
            f"Please migrate to ExecutableWithSocketAgent! For more details see "
            f"https://github.com/RLBot/RLBotJavaExample/wiki/Py4j-Deprecation")

    def run_independently(self, terminate_request_event):
        pass



class EmptyClass:
    pass


def create_enum_object(list, list_name=None, other_attributes=None, attribute_object=None):
    """
    Creates an enum object + adds other attributes.
    :param list: A string list of enum values in order, The first value gets assigned to zero.
    :param list_name: the name of the list if that is also wanting to be stored
    :param other_attributes:  This is a list of tuples.
        The first item in the tuple is the name of the field, the second is the value
    :param attribute_object: An object we should get values for
    :return:
    """
    result = EmptyClass()
    for i in range(len(list)):
        setattr(result, list[i], i if attribute_object is None else getattr(attribute_object, list[i]))
    if list_name is not None:
        setattr(result, list_name, list)
    if other_attributes is not None:
        for attribute_tuple in other_attributes:
            setattr(result, attribute_tuple[0], attribute_tuple[1])
    return result



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// A player that Rocket League treats as human, e.g. has a dedicated camera and can do training mode,
# /// but is actually controlled by a bot.
class PartyMemberBotPlayer(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPartyMemberBotPlayer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PartyMemberBotPlayer()
        x.Init(buf, n + offset)
        return x

    # PartyMemberBotPlayer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

def PartyMemberBotPlayerStart(builder): builder.StartObject(0)
def PartyMemberBotPlayerEnd(builder): return builder.EndObject()



import ctypes
import os
import sys
import time
import platform
from logging import DEBUG, WARNING
from socket import timeout

from rlbot.socket.socket_manager import SocketRelay

import flatbuffers
from flatbuffers import Builder
from rlbot.messages.flat.MatchSettings import MatchSettings as MatchSettingsPacket
from rlbot.messages.flat.BallPrediction import BallPrediction as BallPredictionPacket
from rlbot.messages.flat.FieldInfo import FieldInfo
from rlbot.messages.flat.QuickChatMessages import QuickChatMessages
from rlbot.messages.flat.GameTickPacket import GameTickPacket as Gtp
from rlbot.utils.file_util import get_python_root
from rlbot.utils.game_state_util import GameState, CarState, Physics, Vector3
from rlbot.utils.packetanalysis.valid_packet_detector import ValidPacketDetector
from rlbot.utils.rendering.rendering_manager import RenderingManager
from rlbot.utils.rlbot_exception import get_exception_from_error_code, EmptyDllResponse
from rlbot.utils.structures import game_data_struct
from rlbot.utils.structures.ball_prediction_struct import BallPrediction
from rlbot.utils.structures.bot_input_struct import PlayerInput
from rlbot.utils.structures.game_data_struct import GameTickPacket, ByteBuffer, FieldInfoPacket
from rlbot.utils.structures.game_status import RLBotCoreStatus
from rlbot.utils.structures.rigid_body_struct import RigidBodyTick
from rlbot.utils.structures.start_match_structures import MatchSettings

if platform.system() == 'Windows':
    dll_name_64 = 'RLBot_Core_Interface.dll'
    dll_name_32 = 'RLBot_Core_Interface_32.dll'
elif platform.system() == 'Linux':
    dll_name_64 = 'libRLBotInterface.so'
    dll_name_32 = 'libRLBotInterface32.so'
elif platform.system() == 'Darwin':
    dll_name_64 = 'libRLBotInterface.dylib'
    dll_name_32 = 'libRLBotInterface32.dylib'

def wrap_callback(callback_func):
    def caller(id, status):
        callback_func(id, status)

    return caller


def get_dll_location():
    return os.path.join(get_dll_directory(), dll_name_64)


def get_dll_32_location():
    return os.path.join(get_dll_directory(), dll_name_32)


def is_32_bit_python():
    return sys.maxsize <= 2**32


def get_dll_directory():
    return os.path.join(get_python_root(), 'rlbot', 'dll')


class GameInterface:
    game = None
    start_match_configuration: MatchSettings = None
    match_config: 'MatchConfig' = None
    start_match_flatbuffer: Builder = None
    game_status_callback_type = None
    callback_func = None
    extension = None

    def __init__(self, logger):
        self.logger = logger
        self.dll_path = get_dll_32_location() if is_32_bit_python() else get_dll_location()
        self.renderer = RenderingManager()
        # wait for the dll to load

    def setup_function_types(self):
        self.wait_until_loaded()
        # update live data packet
        func = self.game.UpdateLiveDataPacket
        func.argtypes = [ctypes.POINTER(GameTickPacket)]
        func.restype = ctypes.c_int

        func = self.game.UpdateFieldInfo
        func.argtypes = [ctypes.POINTER(FieldInfoPacket)]
        func.restype = ctypes.c_int

        func = self.game.UpdateLiveDataPacketFlatbuffer
        func.argtypes = []
        func.restype = ByteBuffer

        func = self.game.UpdateRigidBodyTick
        func.argtypes = [ctypes.POINTER(RigidBodyTick)]
        func.restype = ctypes.c_int

        func = self.game.UpdateFieldInfoFlatbuffer
        func.argtypes = []
        func.restype = ByteBuffer

        func = self.game.GetBallPredictionStruct
        func.argtypes = [ctypes.POINTER(BallPrediction)]
        func.restype = ctypes.c_int

        func = self.game.GetBallPrediction
        func.argtypes = []
        func.restype = ByteBuffer

        # start match
        func = self.game.StartMatch
        func.argtypes = [MatchSettings]
        func.restype = ctypes.c_int

        func = self.game.StartMatchFlatbuffer
        func.argtypes = [ctypes.c_void_p, ctypes.c_int]
        func.restype = ctypes.c_int

        # update player input
        func = self.game.UpdatePlayerInput
        func.argtypes = [PlayerInput, ctypes.c_int]
        func.restype = ctypes.c_int

        # update player input
        func = self.game.UpdatePlayerInputFlatbuffer
        func.argtypes = [ctypes.c_void_p, ctypes.c_int]
        func.restype = ctypes.c_int

        # send chat
        func = self.game.SendChat
        func.argtypes = [ctypes.c_uint, ctypes.c_int, ctypes.c_bool, self.game_status_callback_type, ctypes.c_void_p]
        func.restype = ctypes.c_int

        func = self.game.SendQuickChat
        func.argtypes = [ctypes.c_void_p, ctypes.c_int]
        func.restype = ctypes.c_int

        func = self.game.ReceiveChat
        func.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
        func.restype = ByteBuffer

        # set game state
        func = self.game.SetGameState
        func.argtypes = [ctypes.c_void_p, ctypes.c_int]
        func.restype = ctypes.c_int

        func = self.game.GetMatchSettings
        func.argtypes = []
        func.restype = ByteBuffer

        func = self.game.FreshLiveDataPacket
        func.argtypes = [ctypes.POINTER(GameTickPacket), ctypes.c_int, ctypes.c_int]
        func.restype = ctypes.c_int

        func = self.game.FreshLiveDataPacketFlatbuffer
        func.argtypes = [ctypes.c_int, ctypes.c_int]
        func.restype = ByteBuffer

        if hasattr(self.game, 'StartTcpCommunication'):
            # Currently, sockets are only working on Windows and Mac, not linux.
            # Linux has older binaries which do not include this function.
            func = self.game.StartTcpCommunication
            func.argtypes = [ctypes.c_int, ctypes.c_bool, ctypes.c_bool, ctypes.c_bool]
            func.restype = ctypes.c_int

        self.renderer.setup_function_types(self.game)
        self.logger.debug('game interface functions are setup')

        # free the memory at the given pointer
        func = self.game.Free
        func.argtypes = [ctypes.c_void_p]

    def update_live_data_packet(self, game_tick_packet: GameTickPacket):
        rlbot_status = self.game.UpdateLiveDataPacket(game_tick_packet)
        self.game_status(None, rlbot_status)
        return game_tick_packet

    def fresh_live_data_packet(self, game_tick_packet: GameTickPacket, timeout_millis: int, key: int):
        rlbot_status = self.game.FreshLiveDataPacket(game_tick_packet, timeout_millis, key)
        self.game_status(None, rlbot_status)
        return game_tick_packet

    def update_field_info_packet(self, field_info_packet: FieldInfoPacket):
        rlbot_status = self.game.UpdateFieldInfo(field_info_packet)
        self.game_status(None, rlbot_status)
        return field_info_packet

    def update_match_data_packet(self):
        pass

    def start_match(self):
        socket_relay = SocketRelay()
        match_config = self.match_config

        def connect_handler():
            socket_relay.send_match_config(match_config)
            socket_relay.disconnect()

        socket_relay.on_connect_handlers.append(connect_handler)
        try:
            socket_relay.connect_and_run(False, False, False)
        except timeout:
            raise TimeoutError("Took too long to connect to RLBot.exe!")


    def update_player_input(self, player_input, index):
        rlbot_status = self.game.UpdatePlayerInput(player_input, index)
        self.game_status(None, rlbot_status, WARNING)

    def send_chat(self, index, team_only, message_details):
        rlbot_status = self.game.SendChat(message_details, index, team_only, self.create_status_callback(), None)
        self.game_status(None, rlbot_status)

    def send_chat_flat(self, chat_message_builder):
        buf = chat_message_builder.Output()
        rlbot_status = self.game.SendQuickChat(bytes(buf), len(buf))
        self.game_status(None, rlbot_status)
        return rlbot_status

    def receive_chat(self, index, team, last_message_index) -> QuickChatMessages:
        byte_buffer = self.game.ReceiveChat(index, team, last_message_index)
        if byte_buffer.size >= 4:  # GetRootAsGameTickPacket gets angry if the size is less than 4
            # We're counting on this copying the data over to a new memory location so that the original
            # pointer can be freed safely.
            proto_string = ctypes.string_at(byte_buffer.ptr, byte_buffer.size)
            self.game.Free(byte_buffer.ptr)  # Avoid a memory leak
            return QuickChatMessages.GetRootAsQuickChatMessages(proto_string, 0)
        else:
            raise EmptyDllResponse()

    def create_callback(self):
        return

    def game_status(self, id, rlbot_status, level=DEBUG):
        if rlbot_status != RLBotCoreStatus.Success and rlbot_status != RLBotCoreStatus.BufferOverfilled:
            if 0 <= rlbot_status < len(RLBotCoreStatus.status_list):
                self.logger.log(level, "bad status %s", RLBotCoreStatus.status_list[rlbot_status])
            else:
                self.logger.log(level, f"bad status {rlbot_status}")

    def wait_until_loaded(self):
        for i in range(0, 120):
            self.game.IsInitialized.restype = ctypes.c_bool
            is_loaded = self.game.IsInitialized()
            if is_loaded:
                self.logger.info('DLL is initialized!')
                return
            else:
                time.sleep(1)
        raise TimeoutError("RLBot took too long to initialize! Was Rocket League started with the -rlbot flag? "
                           "If you're not sure, close Rocket League and let us open it for you next time!")

    def wait_until_ready_to_communicate(self):
        if hasattr(self.game, 'IsReadyForCommunication'):
            for i in range(0, 120):
                self.game.IsReadyForCommunication.restype = ctypes.c_bool
                is_loaded = self.game.IsReadyForCommunication()
                if is_loaded:
                    self.logger.info('DLL is ready for comms!')
                    return
                else:
                    self.logger.debug('Waiting until DLL is ready to communicate...')
                    time.sleep(1)
            raise TimeoutError("RLBot took too long to initialize! Was Rocket League started with the -rlbot flag? "
                            "If you're not sure, close Rocket League and let us open it for you next time!")


    def wait_until_valid_packet(self):
        detector = ValidPacketDetector(self.match_config)
        detector.wait_until_valid_packet()

    def load_interface(self, port=23234, wants_ball_predictions=True, wants_quick_chat=True, wants_game_messages=False):
        self.game_status_callback_type = ctypes.CFUNCTYPE(None, ctypes.c_uint, ctypes.c_uint)
        self.callback_func = self.game_status_callback_type(wrap_callback(self.game_status))
        self.game = ctypes.CDLL(self.dll_path)
        time.sleep(0.2)
        self.setup_function_types()
        if hasattr(self.game, 'StartTcpCommunication'):
            # Currently, sockets are only working on Windows and Mac, not linux.
            # Linux has older binaries which do not include this function.
            self.logger.info("About to call StartTcp")
            self.game.StartTcpCommunication(port, wants_ball_predictions, wants_quick_chat, wants_game_messages)
            self.wait_until_ready_to_communicate()

    def create_status_callback(self, callback=None):
        """
        Creates a callback for the rlbot status, uses default function if callback is none.
        :param callback:
        :return:
        """
        if callback is None:
            return self.callback_func

        def safe_wrapper(id, rlbotstatsus):
            callback(rlbotstatsus)

        return self.game_status_callback_type(wrap_callback(safe_wrapper))

    def set_extension(self, extension):
        self.game_status_callback_type(wrap_callback(self.game_status))
        self.extension = extension

    def update_player_input_flat(self, player_input_builder):
        buf = player_input_builder.Output()
        rlbot_status = self.game.UpdatePlayerInputFlatbuffer(bytes(buf), len(buf))
        self.game_status(None, rlbot_status, WARNING)

    def set_game_state(self, game_state: GameState) -> None:
        builder = flatbuffers.Builder(0)
        game_state_offset = game_state.convert_to_flat(builder)
        if game_state_offset is None:
            return  # There are no values to be set, so just skip it
        builder.Finish(game_state_offset)
        buf = builder.Output()
        rlbot_status = self.game.SetGameState(bytes(buf), len(buf))
        self.game_status(None, rlbot_status)

    def get_live_data_flat_binary(self):
        """
        Gets the live data packet in flatbuffer binary format. You'll need to do something like
        GameTickPacket.GetRootAsGameTickPacket(binary, 0) to get the data out.

        This is a temporary method designed to keep the integration test working. It returns the raw bytes
        of the flatbuffer so that it can be stored in a file. We can get rid of this once we have a first-class
        data recorder that lives inside the core dll.
        """
        byte_buffer = self.game.UpdateLiveDataPacketFlatbuffer()
        return self._process_live_flatbuffer(byte_buffer)

    def get_fresh_live_data_flat_binary(self, timeout_millis: int, key: int):
        """
        Gets the live data packet in flatbuffer binary format. You'll need to do something like
        GameTickPacket.GetRootAsGameTickPacket(binary, 0) to get the data out.
        This one blocks until a fresh frame is available, or until the timeout has elapsed.
        :param timeout_millis: will give up waiting for a fresh packet and just return a stale one
        after this number of milliseconds.
        :param key: answers the question "fresh from whose perspective?". Freshness will be
        tracked separately for whatever key you pass. Bot index is a reasonable choice.
        """
        byte_buffer = self.game.FreshLiveDataPacketFlatbuffer(timeout_millis, key)
        return self._process_live_flatbuffer(byte_buffer)

    def _process_live_flatbuffer(self, byte_buffer):
        if byte_buffer.size >= 4:  # GetRootAsGameTickPacket gets angry if the size is less than 4
            # We're counting on this copying the data over to a new memory location so that the original
            # pointer can be freed safely.
            proto_string = ctypes.string_at(byte_buffer.ptr, byte_buffer.size)
            self.game.Free(byte_buffer.ptr)  # Avoid a memory leak
            self.game_status(None, RLBotCoreStatus.Success)
            return proto_string

    def update_rigid_body_tick(self, rigid_body_tick: RigidBodyTick):
        """Get the most recent state of the physics engine."""
        rlbot_status = self.game.UpdateRigidBodyTick(rigid_body_tick)
        self.game_status(None, rlbot_status)
        return rigid_body_tick

    def get_field_info(self) -> FieldInfo:
        """
        Gets the field information from the interface.
        :return: The field information
        """
        byte_buffer = self.game.UpdateFieldInfoFlatbuffer()

        if byte_buffer.size >= 4:  # GetRootAsGameTickPacket gets angry if the size is less than 4
            # We're counting on this copying the data over to a new memory location so that the original
            # pointer can be freed safely.
            proto_string = ctypes.string_at(byte_buffer.ptr, byte_buffer.size)
            self.game.Free(byte_buffer.ptr)  # Avoid a memory leak
            self.game_status(None, RLBotCoreStatus.Success)
            return FieldInfo.GetRootAsFieldInfo(proto_string, 0)

    def update_ball_prediction(self, ball_prediction: BallPrediction):
        rlbot_status = self.game.GetBallPredictionStruct(ball_prediction)
        self.game_status(None, rlbot_status)
        return ball_prediction

    def get_ball_prediction(self) -> BallPredictionPacket:
        """
        Gets the latest ball prediction available in shared memory. Only works if BallPrediction.exe is running.
        """
        byte_buffer = self.game.GetBallPrediction()

        if byte_buffer.size >= 4:  # GetRootAsGameTickPacket gets angry if the size is less than 4
            # We're counting on this copying the data over to a new memory location so that the original
            # pointer can be freed safely.
            proto_string = ctypes.string_at(byte_buffer.ptr, byte_buffer.size)
            self.game.Free(byte_buffer.ptr)  # Avoid a memory leak
            self.game_status(None, RLBotCoreStatus.Success)
            return BallPredictionPacket.GetRootAsBallPrediction(proto_string, 0)

    def get_match_settings(self) -> MatchSettings:
        """
        Gets the current match settings in flatbuffer format. Useful for determining map, game mode, mutator settings,
        etc.
        """
        byte_buffer = self.game.GetMatchSettings()

        if byte_buffer.size >= 4:  # GetRootAsGameTickPacket gets angry if the size is less than 4
            # We're counting on this copying the data over to a new memory location so that the original
            # pointer can be freed safely.
            proto_string = ctypes.string_at(byte_buffer.ptr, byte_buffer.size)
            self.game.Free(byte_buffer.ptr)  # Avoid a memory leak
            self.game_status(None, RLBotCoreStatus.Success)
            return MatchSettingsPacket.GetRootAsMatchSettings(proto_string, 0)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Contains all rigid body state information.
class RigidBodyTick(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRigidBodyTick(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RigidBodyTick()
        x.Init(buf, n + offset)
        return x

    # RigidBodyTick
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RigidBodyTick
    def Ball(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .BallRigidBodyState import BallRigidBodyState
            obj = BallRigidBodyState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RigidBodyTick
    def Players(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .PlayerRigidBodyState import PlayerRigidBodyState
            obj = PlayerRigidBodyState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RigidBodyTick
    def PlayersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def RigidBodyTickStart(builder): builder.StartObject(2)
def RigidBodyTickAddBall(builder, ball): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(ball), 0)
def RigidBodyTickAddPlayers(builder, players): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(players), 0)
def RigidBodyTickStartPlayersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RigidBodyTickEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// The car type, color, and other aspects of the player's appearance.
# /// See https://github.com/RLBot/RLBot/wiki/Bot-Customization
class PlayerLoadout(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerLoadout(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerLoadout()
        x.Init(buf, n + offset)
        return x

    # PlayerLoadout
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PlayerLoadout
    def TeamColorId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def CustomColorId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def CarId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def DecalId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def WheelsId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def BoostId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def AntennaId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def HatId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def PaintFinishId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def CustomFinishId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def EngineAudioId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def TrailsId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def GoalExplosionId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerLoadout
    def LoadoutPaint(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .LoadoutPaint import LoadoutPaint
            obj = LoadoutPaint()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// Sets the primary color of the car to the swatch that most closely matches the provided
# /// RGB color value. If set, this overrides teamColorId.
    # PlayerLoadout
    def PrimaryColorLookup(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Color import Color
            obj = Color()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// Sets the secondary color of the car to the swatch that most closely matches the provided
# /// RGB color value. If set, this overrides customColorId.
    # PlayerLoadout
    def SecondaryColorLookup(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Color import Color
            obj = Color()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def PlayerLoadoutStart(builder): builder.StartObject(16)
def PlayerLoadoutAddTeamColorId(builder, teamColorId): builder.PrependInt32Slot(0, teamColorId, 0)
def PlayerLoadoutAddCustomColorId(builder, customColorId): builder.PrependInt32Slot(1, customColorId, 0)
def PlayerLoadoutAddCarId(builder, carId): builder.PrependInt32Slot(2, carId, 0)
def PlayerLoadoutAddDecalId(builder, decalId): builder.PrependInt32Slot(3, decalId, 0)
def PlayerLoadoutAddWheelsId(builder, wheelsId): builder.PrependInt32Slot(4, wheelsId, 0)
def PlayerLoadoutAddBoostId(builder, boostId): builder.PrependInt32Slot(5, boostId, 0)
def PlayerLoadoutAddAntennaId(builder, antennaId): builder.PrependInt32Slot(6, antennaId, 0)
def PlayerLoadoutAddHatId(builder, hatId): builder.PrependInt32Slot(7, hatId, 0)
def PlayerLoadoutAddPaintFinishId(builder, paintFinishId): builder.PrependInt32Slot(8, paintFinishId, 0)
def PlayerLoadoutAddCustomFinishId(builder, customFinishId): builder.PrependInt32Slot(9, customFinishId, 0)
def PlayerLoadoutAddEngineAudioId(builder, engineAudioId): builder.PrependInt32Slot(10, engineAudioId, 0)
def PlayerLoadoutAddTrailsId(builder, trailsId): builder.PrependInt32Slot(11, trailsId, 0)
def PlayerLoadoutAddGoalExplosionId(builder, goalExplosionId): builder.PrependInt32Slot(12, goalExplosionId, 0)
def PlayerLoadoutAddLoadoutPaint(builder, loadoutPaint): builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(loadoutPaint), 0)
def PlayerLoadoutAddPrimaryColorLookup(builder, primaryColorLookup): builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(primaryColorLookup), 0)
def PlayerLoadoutAddSecondaryColorLookup(builder, secondaryColorLookup): builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(secondaryColorLookup), 0)
def PlayerLoadoutEnd(builder): return builder.EndObject()



from pathlib import Path
from enum import Enum
from random import randint
from typing import List, Dict

from flatbuffers import Builder
from rlbot.matchconfig.loadout_config import LoadoutConfig
from rlbot.messages.flat.PlayerClass import PlayerClass
from rlbot.parsing.bot_config_bundle import get_bot_config_bundle
from rlbot.parsing.agent_config_parser import load_bot_appearance
from rlbot.parsing.match_settings_config_parser import boost_amount_mutator_types, map_types, game_mode_types, \
    match_length_types, max_score_types, overtime_mutator_types, series_length_mutator_types, game_speed_mutator_types, \
    ball_max_speed_mutator_types, ball_type_mutator_types, ball_weight_mutator_types, ball_size_mutator_types, \
    ball_bounciness_mutator_types, rumble_mutator_types, boost_strength_mutator_types, gravity_mutator_types, \
    demolish_mutator_types, respawn_time_mutator_types, existing_match_behavior_types, game_map_dict
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.structures.start_match_structures import MatchSettings, PlayerConfiguration, MutatorSettings
from rlbot.messages.flat import MatchSettings as MatchSettingsFlat, MutatorSettings as MutatorSettingsFlat, RLBotPlayer, \
    PsyonixBotPlayer, HumanPlayer, PlayerConfiguration as PlayerConfigurationFlat

# We pass messages in flatbuffer format to RLBot.exe. In flatbuffer, a signed int field
# is 32 bit, so it has a max value of 2^31 - 1, in other words 2147483647.
FLATBUFFER_MAX_INT = 2**31 - 1

class Team(Enum):
    BLUE = 0
    ORANGE = 1

class ScriptConfig:
    def __init__(self, config_path):
        self.config_path: str = config_path

class PlayerConfig:
    """
    Represents configuration for a player (could be a bot, a human, etc).
    Encompasses everything, including loadout, behavior, etc.

    Knows how to translate itself into the PlayerConfiguration ctypes class.
    """

    def __init__(self):
        self.bot: bool = None
        self.rlbot_controlled: bool = None
        self.bot_skill: float = None
        self.human_index: int = None
        self.name: str = None
        self.team: int = None
        self.config_path: str = None  # Required only if rlbot_controlled is true
        self.loadout_config: LoadoutConfig = None
        self.spawn_id: int = randint(1, FLATBUFFER_MAX_INT)  # Feel free to override this.

    @staticmethod  # TODO: in Python 3.7 we can remove the quotes from the return type.
    def bot_config(player_config_path: Path, team: Team) -> 'PlayerConfig':
        """
        A function to cover the common case of creating a config for a bot.
        """
        bot_config = PlayerConfig()
        bot_config.bot = True
        bot_config.rlbot_controlled = True
        bot_config.team = team.value
        bot_config.config_path = str(player_config_path.absolute())  # TODO: Refactor to use Path's
        config_bundle = get_bot_config_bundle(bot_config.config_path)
        bot_config.name = config_bundle.name
        bot_config.loadout_config = load_bot_appearance(config_bundle.get_looks_config(), bot_config.team)
        return bot_config

    def write(self, player_configuration: PlayerConfiguration, name_dict: dict):
        player_configuration.bot = self.bot
        player_configuration.rlbot_controlled = self.rlbot_controlled
        player_configuration.bot_skill = self.bot_skill or 0
        player_configuration.human_index = self.human_index or 0
        player_configuration.name = get_sanitized_bot_name(name_dict, self.name)
        player_configuration.team = self.team
        player_configuration.spawn_id = self.spawn_id

        if self.loadout_config:
            self.loadout_config.write(player_configuration)

    def write_to_flatbuffer(self, builder: Builder, name_dict: dict):
        name = builder.CreateString(get_sanitized_bot_name(name_dict, self.name))

        if self.loadout_config:
            loadout = self.loadout_config.write_to_flatbuffer(builder)
            if self.rlbot_controlled:
                variety = PlayerClass.RLBotPlayer
                RLBotPlayer.RLBotPlayerStart(builder)
                player = RLBotPlayer.RLBotPlayerEnd(builder)
            else:
                variety = PlayerClass.PsyonixBotPlayer
                PsyonixBotPlayer.PsyonixBotPlayerStart(builder)
                PsyonixBotPlayer.PsyonixBotPlayerAddBotSkill(builder, self.bot_skill)
                player = PsyonixBotPlayer.PsyonixBotPlayerEnd(builder)
        else:
            loadout = LoadoutConfig().write_to_flatbuffer(builder)
            variety = PlayerClass.HumanPlayer
            HumanPlayer.HumanPlayerStart(builder)
            player = HumanPlayer.HumanPlayerEnd(builder)

        PlayerConfigurationFlat.PlayerConfigurationStart(builder)
        PlayerConfigurationFlat.PlayerConfigurationAddName(builder, name)
        PlayerConfigurationFlat.PlayerConfigurationAddLoadout(builder, loadout)
        PlayerConfigurationFlat.PlayerConfigurationAddTeam(builder, self.team)
        PlayerConfigurationFlat.PlayerConfigurationAddVariety(builder, player)
        PlayerConfigurationFlat.PlayerConfigurationAddVarietyType(builder, variety)
        PlayerConfigurationFlat.PlayerConfigurationAddSpawnId(builder, self.spawn_id)
        return PlayerConfigurationFlat.PlayerConfigurationEnd(builder)

    def has_bot_script(self) -> bool:
        return self.rlbot_controlled

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class EmptyPlayerSlot(PlayerConfig):
    """
    This is useful for keeping a player index empty.
    """
    def write(self, player_configuration: PlayerConfiguration, name_dict: dict):
        # This is a dirty trick. With bot = False and rlbot_controlled = True, we have a "party member bot"
        # which is not supported in RLBot.exe anymore and will just get skipped over. That's what we want.
        player_configuration.bot = False
        player_configuration.rlbot_controlled = True
        player_configuration.spawn_id = -1
        player_configuration.name = ""

    def has_bot_script(self) -> bool:
        return False


def index_or_zero(types, value):
    if value is None:
        return 0
    return types.index(value)


class MutatorConfig:
    """
    Represent mutator configuration, e.g. match length, boost amount, etc.

    Knows how to translate itself into the MutatorSettings ctypes class.
    """

    def __init__(self):
        self.match_length: str = None
        self.max_score: str = None
        self.overtime: str = None
        self.series_length: str = None
        self.game_speed: str = None
        self.ball_max_speed: str = None
        self.ball_type: str = None
        self.ball_weight: str = None
        self.ball_size: str = None
        self.ball_bounciness: str = None
        self.boost_amount: str = None
        self.rumble: str = None
        self.boost_strength: str = None
        self.gravity: str = None
        self.demolish: str = None
        self.respawn_time: str = None

    def write(self, mutator_settings: MutatorSettings):
        mutator_settings.match_length = index_or_zero(match_length_types, self.match_length)
        mutator_settings.max_score = index_or_zero(max_score_types, self.max_score)
        mutator_settings.overtime_option = index_or_zero(overtime_mutator_types, self.overtime)
        mutator_settings.series_length_option = index_or_zero(series_length_mutator_types, self.series_length)
        mutator_settings.game_speed_option = index_or_zero(game_speed_mutator_types, self.game_speed)
        mutator_settings.ball_max_speed_option = index_or_zero(ball_max_speed_mutator_types, self.ball_max_speed)
        mutator_settings.ball_type_option = index_or_zero(ball_type_mutator_types, self.ball_type)
        mutator_settings.ball_weight_option = index_or_zero(ball_weight_mutator_types, self.ball_weight)
        mutator_settings.ball_size_option = index_or_zero(ball_size_mutator_types, self.ball_size)
        mutator_settings.ball_bounciness_option = index_or_zero(ball_bounciness_mutator_types, self.ball_bounciness)
        mutator_settings.boost_amount_option = index_or_zero(boost_amount_mutator_types, self.boost_amount)
        mutator_settings.rumble_option = index_or_zero(rumble_mutator_types, self.rumble)
        mutator_settings.boost_strength_option = index_or_zero(boost_strength_mutator_types, self.boost_strength)
        mutator_settings.gravity_option = index_or_zero(gravity_mutator_types, self.gravity)
        mutator_settings.demolish_option = index_or_zero(demolish_mutator_types, self.demolish)
        mutator_settings.respawn_time_option = index_or_zero(respawn_time_mutator_types, self.respawn_time)

    def write_to_flatbuffer(self, builder: Builder):
        MutatorSettingsFlat.MutatorSettingsStart(builder)
        MutatorSettingsFlat.MutatorSettingsAddMatchLength(builder, index_or_zero(match_length_types, self.match_length))
        MutatorSettingsFlat.MutatorSettingsAddMaxScore(builder, index_or_zero(max_score_types, self.max_score))
        MutatorSettingsFlat.MutatorSettingsAddOvertimeOption(builder, index_or_zero(overtime_mutator_types, self.overtime))
        MutatorSettingsFlat.MutatorSettingsAddSeriesLengthOption(builder, index_or_zero(series_length_mutator_types, self.series_length))
        MutatorSettingsFlat.MutatorSettingsAddGameSpeedOption(builder, index_or_zero(game_speed_mutator_types, self.game_speed))
        MutatorSettingsFlat.MutatorSettingsAddBallMaxSpeedOption(builder, index_or_zero(ball_max_speed_mutator_types, self.ball_max_speed))
        MutatorSettingsFlat.MutatorSettingsAddBallTypeOption(builder, index_or_zero(ball_type_mutator_types, self.ball_type))
        MutatorSettingsFlat.MutatorSettingsAddBallWeightOption(builder, index_or_zero(ball_weight_mutator_types, self.ball_weight))
        MutatorSettingsFlat.MutatorSettingsAddBallSizeOption(builder, index_or_zero(ball_size_mutator_types, self.ball_size))
        MutatorSettingsFlat.MutatorSettingsAddBallBouncinessOption(builder, index_or_zero(ball_bounciness_mutator_types, self.ball_bounciness))
        MutatorSettingsFlat.MutatorSettingsAddBoostOption(builder, index_or_zero(boost_amount_mutator_types, self.boost_amount))
        MutatorSettingsFlat.MutatorSettingsAddRumbleOption(builder, index_or_zero(rumble_mutator_types, self.rumble))
        MutatorSettingsFlat.MutatorSettingsAddBoostStrengthOption(builder, index_or_zero(boost_strength_mutator_types, self.boost_strength))
        MutatorSettingsFlat.MutatorSettingsAddGravityOption(builder, index_or_zero(gravity_mutator_types, self.gravity))
        MutatorSettingsFlat.MutatorSettingsAddDemolishOption(builder, index_or_zero(demolish_mutator_types, self.demolish))
        MutatorSettingsFlat.MutatorSettingsAddRespawnTimeOption(builder, index_or_zero(respawn_time_mutator_types, self.respawn_time))
        return MutatorSettingsFlat.MutatorSettingsEnd(builder)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    @staticmethod
    def from_mutator_settings_flatbuffer(mutator_settings: MutatorSettingsFlat):
        mc = MutatorConfig()
        mc.match_length = match_length_types[mutator_settings.MatchLength()]
        mc.max_score = max_score_types[mutator_settings.MaxScore()]
        mc.overtime = overtime_mutator_types[mutator_settings.OvertimeOption()]
        mc.series_length = series_length_mutator_types[mutator_settings.SeriesLengthOption()]
        mc.game_speed = game_speed_mutator_types[mutator_settings.GameSpeedOption()]
        mc.ball_max_speed = ball_max_speed_mutator_types[mutator_settings.BallMaxSpeedOption()]
        mc.ball_type = ball_type_mutator_types[mutator_settings.BallTypeOption()]
        mc.ball_weight = ball_weight_mutator_types[mutator_settings.BallWeightOption()]
        mc.ball_size = ball_size_mutator_types[mutator_settings.BallSizeOption()]
        mc.ball_bounciness = ball_bounciness_mutator_types[mutator_settings.BallBouncinessOption()]
        mc.boost_amount = boost_amount_mutator_types[mutator_settings.BoostOption()]
        mc.rumble = rumble_mutator_types[mutator_settings.RumbleOption()]
        mc.boost_strength = boost_strength_mutator_types[mutator_settings.BoostStrengthOption()]
        mc.gravity = gravity_mutator_types[mutator_settings.GravityOption()]
        mc.demolish = demolish_mutator_types[mutator_settings.DemolishOption()]
        mc.respawn_time = respawn_time_mutator_types[mutator_settings.RespawnTimeOption()]
        return mc


class ExtensionConfig:
    def __init__(self):
        self.python_file_path: str = None

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class MatchConfig:
    """
    Represents configuration for an entire match. Includes player config and mutators.
    """

    def __init__(self):
        self.player_configs: List[PlayerConfig] = []
        self.game_mode: str = None
        self.game_map: str = None
        self.skip_replays: bool = False
        self.instant_start: bool = False
        self.mutators: MutatorConfig = None
        self.extension_config: ExtensionConfig = None
        self.existing_match_behavior: str = None
        self.enable_lockstep: bool = False
        self.networking_role: str = None
        self.network_address: str = None
        self.enable_rendering: bool = False
        self.enable_state_setting: bool = False
        self.auto_save_replay: bool = False
        self.script_configs: List[ScriptConfig] = []
        self.logger = get_logger('match_config')

    @property
    def num_players(self):
        return len(self.player_configs)

    # TODO: retire the MatchSettings object and just build a flatbuffer here instead.
    def create_match_settings(self) -> MatchSettings:
        match_settings = MatchSettings()
        name_dict = {}
        for index, player_config in enumerate(self.player_configs):
            player_config.write(match_settings.player_configuration[index], name_dict)
        match_settings.num_players = self.num_players
        match_settings.game_mode = game_mode_types.index(self.game_mode)
        match_settings.game_map = map_types.index(self.game_map)
        match_settings.skip_replays = self.skip_replays
        match_settings.instant_start = self.instant_start
        match_settings.existing_match_behavior = index_or_zero(existing_match_behavior_types, self.existing_match_behavior)
        match_settings.enable_lockstep = self.enable_lockstep
        match_settings.enable_rendering = self.enable_rendering
        match_settings.enable_state_setting = self.enable_state_setting
        match_settings.auto_save_replay = self.auto_save_replay

        self.mutators.write(match_settings.mutator_settings)

        return match_settings

    def create_flatbuffer(self) -> Builder:
        builder = Builder(1000)
        name_dict = {}
        player_config_offsets = [pc.write_to_flatbuffer(builder, name_dict) for pc in self.player_configs]
        MatchSettingsFlat.MatchSettingsStartPlayerConfigurationsVector(builder, len(player_config_offsets))
        for i in reversed(range(0, len(player_config_offsets))):
            builder.PrependUOffsetTRelative(player_config_offsets[i])
        player_list_offset = builder.EndVector(len(player_config_offsets))
        mutator_settings_offset = self.mutators.write_to_flatbuffer(builder)

        if self.game_map in game_map_dict:
            upk = game_map_dict[self.game_map]
            map_index = map_types.index(self.game_map)
        else:
            self.logger.info(f"Did not recognize {self.game_map}, hoping it's new or a custom map!")
            upk = self.game_map
            map_index = -1
        upk_offset = builder.CreateString(upk)

        MatchSettingsFlat.MatchSettingsStart(builder)
        MatchSettingsFlat.MatchSettingsAddPlayerConfigurations(builder, player_list_offset)
        MatchSettingsFlat.MatchSettingsAddGameMode(builder, game_mode_types.index(self.game_mode))
        MatchSettingsFlat.MatchSettingsAddGameMap(builder, map_index)
        MatchSettingsFlat.MatchSettingsAddGameMapUpk(builder, upk_offset)
        MatchSettingsFlat.MatchSettingsAddSkipReplays(builder, self.skip_replays)
        MatchSettingsFlat.MatchSettingsAddInstantStart(builder, self.instant_start)
        MatchSettingsFlat.MatchSettingsAddMutatorSettings(builder, mutator_settings_offset)
        MatchSettingsFlat.MatchSettingsAddExistingMatchBehavior(builder, index_or_zero(existing_match_behavior_types, self.existing_match_behavior))
        MatchSettingsFlat.MatchSettingsAddEnableLockstep(builder, self.enable_lockstep)
        MatchSettingsFlat.MatchSettingsAddEnableRendering(builder, self.enable_rendering)
        MatchSettingsFlat.MatchSettingsAddEnableStateSetting(builder, self.enable_state_setting)
        MatchSettingsFlat.MatchSettingsAddAutoSaveReplay(builder, self.auto_save_replay)
        ms_offset = MatchSettingsFlat.MatchSettingsEnd(builder)
        builder.Finish(ms_offset)
        return builder

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    @staticmethod
    def from_match_settings_flatbuffer(match_settings: MatchSettingsFlat):
        mc = MatchConfig()
        mc.game_mode = game_mode_types[match_settings.GameMode()]
        mc.game_map = map_types[match_settings.GameMap()]
        mc.mutators = MutatorConfig.from_mutator_settings_flatbuffer(match_settings.MutatorSettings())
        # TODO: player configs
        return mc


def get_sanitized_bot_name(dict: Dict[str, int], name: str) -> str:
    """
    Cut off at 31 characters and handle duplicates.
    :param dict: Holds the list of names for duplicates
    :param name: The name that is being sanitized
    :return: A sanitized version of the name
    """

    # This doesn't work someimtes in continue_and_spawn because it doesn't understand the names already in the match
    # which may be kept if the spawn IDs match. In that case it's the caller's responsibility to figure it out upstream.

    name = name[:31]
    base_name = name
    count = 2
    while name in dict:
        name = f'{base_name[:27]} ({count})'  # Truncate at 27 because we can have up to '(10)' appended
        count += 1
    dict[name] = 1
    return name



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class RotatorPartial(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRotatorPartial(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RotatorPartial()
        x.Init(buf, n + offset)
        return x

    # RotatorPartial
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RotatorPartial
    def Pitch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RotatorPartial
    def Yaw(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RotatorPartial
    def Roll(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def RotatorPartialStart(builder): builder.StartObject(3)
def RotatorPartialAddPitch(builder, pitch): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(pitch), 0)
def RotatorPartialAddYaw(builder, yaw): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(yaw), 0)
def RotatorPartialAddRoll(builder, roll): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(roll), 0)
def RotatorPartialEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DesiredGameState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDesiredGameState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DesiredGameState()
        x.Init(buf, n + offset)
        return x

    # DesiredGameState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DesiredGameState
    def BallState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .DesiredBallState import DesiredBallState
            obj = DesiredBallState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameState
    def CarStates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .DesiredCarState import DesiredCarState
            obj = DesiredCarState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameState
    def CarStatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # DesiredGameState
    def BoostStates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .DesiredBoostState import DesiredBoostState
            obj = DesiredBoostState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameState
    def BoostStatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # DesiredGameState
    def GameInfoState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .DesiredGameInfoState import DesiredGameInfoState
            obj = DesiredGameInfoState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameState
    def ConsoleCommands(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .ConsoleCommand import ConsoleCommand
            obj = ConsoleCommand()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameState
    def ConsoleCommandsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def DesiredGameStateStart(builder): builder.StartObject(5)
def DesiredGameStateAddBallState(builder, ballState): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(ballState), 0)
def DesiredGameStateAddCarStates(builder, carStates): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(carStates), 0)
def DesiredGameStateStartCarStatesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def DesiredGameStateAddBoostStates(builder, boostStates): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(boostStates), 0)
def DesiredGameStateStartBoostStatesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def DesiredGameStateAddGameInfoState(builder, gameInfoState): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(gameInfoState), 0)
def DesiredGameStateAddConsoleCommands(builder, consoleCommands): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(consoleCommands), 0)
def DesiredGameStateStartConsoleCommandsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def DesiredGameStateEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// A minimal version of the ball, useful when bandwidth needs to be conserved.
class TinyBall(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTinyBall(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TinyBall()
        x.Init(buf, n + offset)
        return x

    # TinyBall
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TinyBall
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TinyBall
    def Velocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def TinyBallStart(builder): builder.StartObject(2)
def TinyBallAddLocation(builder, location): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def TinyBallAddVelocity(builder, velocity): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(velocity), 0)
def TinyBallEnd(builder): return builder.EndObject()



import os
import socket
import time

import psutil
from rlbot.agents.base_independent_agent import BaseIndependentAgent
from rlbot.botmanager.helper_process_request import HelperProcessRequest
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.structures import game_interface
from rlbot.utils.structures.game_data_struct import GameTickPacket
from rlbot.utils.structures.game_interface import GameInterface


class ExecutableWithSocketAgent(BaseIndependentAgent):

    def __init__(self, name, team, index):
        super().__init__(name, team, index)
        self.logger = get_logger('ExeSocket' + str(self.index))
        self.is_retired = False
        self.executable_path = None
        self.game_interface = GameInterface(self.logger)
        self.game_tick_packet = GameTickPacket()
        self.spawn_id_seen = False

    def run_independently(self, terminate_request_event):
        self.game_interface.load_interface()

        while not terminate_request_event.is_set():

            self.game_interface.update_live_data_packet(self.game_tick_packet)
            packet_spawn_id = self.game_tick_packet.game_cars[self.index].spawn_id
            if self.spawn_id_seen:
                if packet_spawn_id != self.spawn_id:
                    break  # This will cause the bot to retire.
            elif packet_spawn_id == self.spawn_id and self.game_tick_packet.game_info.is_round_active:
                self.spawn_id_seen = True

            # Continuously make sure the the bot is registered.
            # These functions can be called repeatedly without any bad effects.
            # This is useful for re-engaging the socket server if it gets restarted during development.
            message = self.build_add_command()
            self.send_command(message)
            time.sleep(1)

    def get_helper_process_request(self):
        if self.is_executable_configured():
            return HelperProcessRequest(python_file_path=None, key=__file__ + str(self.get_port()),
                                        executable=self.executable_path, exe_args=[str(self.get_port())],
                                        current_working_directory=os.path.dirname(self.executable_path))
        return None

    def retire(self):
        message = self.build_retire_command()
        self.logger.info(f"Sending retire message for {self.name}")
        self.send_command(message)
        self.is_retired = True

    def build_add_command(self) -> str:
        return f"add\n{self.name}\n{self.team}\n{self.index}\n{game_interface.get_dll_directory()}"

    def build_retire_command(self) -> str:
        return f"remove\n{self.index}"

    def send_command(self, message):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(4)
            s.connect(("127.0.0.1", self.get_port()))
            s.send(bytes(message, "ASCII"))
            s.close()
            return True
        except ConnectionRefusedError:
            self.logger.warn("Could not connect to server!")
            return False

    def is_executable_configured(self):
        return self.executable_path is not None and os.path.isfile(self.executable_path)

    def get_extra_pids(self):
        """
        Gets the list of process ids that should be marked as high priority.
        :return: A list of process ids that are used by this bot in addition to the ones inside the python process.
        """

        if self.is_executable_configured():
            # The helper process will start the exe and report the PID. Nothing to do here.
            return []

        while not self.is_retired:
            for proc in psutil.process_iter():
                for conn in proc.connections():
                    if conn.laddr.port == self.get_port():
                        self.logger.debug(f'server for {self.name} appears to have pid {proc.pid}')
                        return [proc.pid]
            time.sleep(1)
            if self.executable_path is None:
                self.logger.info(
                    "Can't auto-start because no executable is configured. Please start manually!")
            else:
                self.logger.info(f"Can't auto-start because {self.executable_path} is not found. "
                                 "Please start manually!")

    def get_port(self) -> int:
        raise NotImplementedError



import time
from collections import namedtuple, defaultdict
from typing import Optional, List
import random
import string
from queue import Empty

from rlbot.matchcomms.client import MatchcommsClient
from rlbot.matchcomms.shared import JSON

MESSAGE_KEY = 'reply_id'

ReplyId = str  # type alias
ID_CHARACTER_OPTIONS = string.ascii_uppercase + string.ascii_lowercase + string.digits


def add_reply_field(outgoing_msg: JSON) -> ReplyId:
    """
    Mutates outgoing_msg to gain a reply_id and returns the reply_id.
    """
    assert not hasattr(outgoing_msg, MESSAGE_KEY), "Message contents conflict with reply MESSAGE_KEY."
    reply_id = generate_random_reply_id()
    outgoing_msg[MESSAGE_KEY] = reply_id
    return reply_id


def reply_to(matchcomms: MatchcommsClient, incoming_msg: JSON, outgoing_msg: JSON = None):
    """
    Mutates outgoing_msg (empty message by default) to gain the reply_id of
    the incoming_msg and broadcasts it.
    """
    if outgoing_msg is None:
        outgoing_msg = {}
    assert not hasattr(outgoing_msg, MESSAGE_KEY), "Message contents conflict with reply MESSAGE_KEY."
    outgoing_msg[MESSAGE_KEY] = get_reply_id(incoming_msg)
    matchcomms.outgoing_broadcast.put_nowait(outgoing_msg)

def is_reply_msg(msg: JSON) -> bool:
    return bool(get_reply_id(msg))


def get_reply_id(msg: JSON) -> Optional[ReplyId]:
    if type(msg) != dict: return None
    return msg.get(MESSAGE_KEY, None)

def generate_random_reply_id():
    return ''.join(random.choice(ID_CHARACTER_OPTIONS) for _ in range(10))



class ResponseDeadlineExceeded(Exception):
    pass

RecievedMessages = namedtuple('RecievedMessages', 'replies unhandled_messages')
def send_and_wait_for_replies(
        matchcomms: MatchcommsClient, outgoing_msgs: List[JSON],
        timeout_seconds=0.1, num_retries=2) -> RecievedMessages:
    """
    Sends all outgoing_msgs then blocks until all of them got a reply.
    Raises a ResponseDeadlineExceeded exception if the number of reties htimeout is reached.
    """
    unhandled_messages = []
    reply_ids = set(add_reply_field(msg) for msg in outgoing_msgs)
    replies = defaultdict(list) # ReplyId -> List[JSON]
    for i in range(num_retries + 1):
        # (re)-send messages that have not gotten a reply yet
        for msg in outgoing_msgs:
            if get_reply_id(msg) not in replies:
                matchcomms.outgoing_broadcast.put_nowait(msg)

        # Poll for replies until the timeout.
        begin = time.time()
        while time.time() - begin < timeout_seconds:
            try:
                msg = matchcomms.incoming_broadcast.get_nowait()
            except Empty:
                time.sleep(0.002)
                continue
            reply_id = get_reply_id(msg)
            if reply_id not in reply_ids:
                unhandled_messages.append(msg)
                continue
            replies[reply_id].append(msg)
            if len(replies) == len(outgoing_msgs):
                return RecievedMessages(replies=replies, unhandled_messages=unhandled_messages)

    unreplied_messages = [msg for msg in outgoing_msgs if get_reply_id(msg) not in replies]
    raise ResponseDeadlineExceeded(f'Did not get responses to these messages: {unreplied_messages}')




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Notification when the local player is spectating another player.
class PlayerSpectate(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerSpectate(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerSpectate()
        x.Init(buf, n + offset)
        return x

    # PlayerSpectate
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// index of the player that is being spectated. Will be -1 if not spectating anyone.
    # PlayerSpectate
    def PlayerIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def PlayerSpectateStart(builder): builder.StartObject(1)
def PlayerSpectateAddPlayerIndex(builder, playerIndex): builder.PrependInt32Slot(0, playerIndex, 0)
def PlayerSpectateEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// We have some very small messages that are only a few bytes but potentially sent at high frequency.
# /// Bundle them into a packet to reduce the overhead of sending data over TCP.
class MessagePacket(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMessagePacket(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MessagePacket()
        x.Init(buf, n + offset)
        return x

    # MessagePacket
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MessagePacket
    def Messages(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .GameMessageWrapper import GameMessageWrapper
            obj = GameMessageWrapper()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # MessagePacket
    def MessagesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MessagePacket
    def GameSeconds(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MessagePacket
    def FrameNum(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def MessagePacketStart(builder): builder.StartObject(3)
def MessagePacketAddMessages(builder, messages): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(messages), 0)
def MessagePacketStartMessagesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def MessagePacketAddGameSeconds(builder, gameSeconds): builder.PrependFloat32Slot(1, gameSeconds, 0.0)
def MessagePacketAddFrameNum(builder, frameNum): builder.PrependInt32Slot(2, frameNum, 0)
def MessagePacketEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class Bool(object):
    __slots__ = ['_tab']

    # Bool
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Bool
    def Val(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))

def CreateBool(builder, val):
    builder.Prep(1, 1)
    builder.PrependBool(val)
    return builder.Offset()



from datetime import date
# data in the form of: ID,FullName,AssetPackageName,AsciiLabel,PackID,Quality,UnlockMethod

dump_file = input('Data dump (input) file name?->')
data = open(dump_file, 'r')

Bodies, Wheels, Boosts, Antennas, Decals, Toppers, Trails, Goal_Explosions, Paints, Banners, Engine_Audio = [
], [], [], [], [], [], [], [], [], [], []

car_dict = {'backfire': 'Backfire', 'force': 'Breakout', 'octane': 'Octane', 'orion': 'Paladin',
            'rhino': 'Road Hog', 'spark': 'Gizmo', 'torch': 'X-Devil', 'torment': 'Hotshot', 'vanquish': 'Merc',
            'venom': 'Venom', 'import': 'Takumi', 'musclecar': 'Dominus', 'zippy': 'Zippy', 'scarab': 'Scarab',
            'interceptor': 'Ripper', 'wastelandtruck': 'Grog', 'musclecar2': 'Dominus GT', 'torch2': 'X-Devil Mk2',
            'neocar': 'Masamune', 'rhino2': 'Road Hog XL', 'takumi': 'Takumi RX-T', 'aftershock': 'Aftershock',
            'neobike': 'Esper', 'marauder': 'Marauder', 'carcar': 'Breakout Type-S', 'number6': 'Proteus',
            'cannonboy': 'Triton', 'o2': 'Octane ZSR', 'gilliam': 'Vulcan', 'bone': 'Bone Shaker',
            'scallop': 'Twin Mill III', 'endo': 'Endo', 'charged': 'Ice Charger', 'flatbread': 'Mantis',
            'pumpernickel': 'Centio V17', 'focaccia': 'Animus GP', 'challah': "'70 Dodge Charger R/T",
            'melonpan': "'99 Nissan Skyline GT-R R34", 'levain': 'Werewolf', 'greycar': 'DeLorean Time Machine',
            'darkcar': "'16 Batmobile", 'takumi_ii': 'Takumi RX-T', 'sourdough': 'Jäger 619 RS',
            'multigrain': 'Imperator DT5', 'berry': 'The Dark Knight Rises Tumbler', 'eggplant': "'89 Batmobile",
            'Universal Decal': 'Universal Decal'}

categories = {
    "Bodies": [Bodies, ['body']],
    "Decals": [Decals, ['skin']],
    "Wheels": [Wheels, ['wheel']],
    "Boosts": [Boosts, ['boost']],
    "Antennas": [Antennas, ['flag', 'antenna', 'pennant', '.at']],
    "Toppers": [Toppers, ['hat', 'crown']],
    "Trails": [Trails, ['ss']],
    "Goal_Explosions": [Goal_Explosions, ['explosion']],
    "Paints": [Paints, ['paintfinish']],
    "Banners": [Banners, ['playerbanner']],
    "Engine_Audio": [Engine_Audio, ['engineaudio']]
}

for line in data:
    line = line.lower()
    line_parts = line.split(',')
    ID = line_parts[0]
    FullName = line_parts[1]
    AssetPackageName = line_parts[2]
    AsciiLabel = line_parts[3]
    PackID = line_parts[4]
    Quality = line_parts[5]
    UnlockMethod = line_parts[6]

    # if it is platform specific, logo for season play, crate, bots, keys, packs
    if (ID == '0' or AssetPackageName == 'seasonlogos' or AssetPackageName == 'itemcontainer' or
            AssetPackageName == 'bots' or 'key' in AssetPackageName or 'pack' in AssetPackageName):
        continue

    else:
        filtered = False
        for cat in categories:
            filters = categories[cat][1]
            for filter in filters:
                if filter in FullName and not filtered:
                    name = ('{:4} - {}'.format(ID, AsciiLabel))
                    categories[cat][0].append(name)
                    filtered = True
    if not filtered:
        print('unassigned:\n', line_parts)


def write_file(categories):
    file_name = input('Clean file (output) name(with "".txt")->')
    clean_file = open(file_name, 'w')

    day = date.today()
    today = day.strftime("%d/%m/%y")
    title = "# List of Item ID's in game as of " + today + "\n\n"
    clean_file.write(title)

    clean_file.write('### ID - AsciiLabel')
    for cat in categories:
        write = '\n\n### ' + cat + '\n'
        clean_file.write(write)
        clean_file.write('```\n')
        for item in categories[cat][0]:
            clean_file.write(item)
            clean_file.write('\n')
        clean_file.write('```')
    clean_file.close()


write_file(categories)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class Physics(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPhysics(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Physics()
        x.Init(buf, n + offset)
        return x

    # Physics
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Physics
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Physics
    def Rotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Rotator import Rotator
            obj = Rotator()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Physics
    def Velocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Physics
    def AngularVelocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def PhysicsStart(builder): builder.StartObject(4)
def PhysicsAddLocation(builder, location): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def PhysicsAddRotation(builder, rotation): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(rotation), 0)
def PhysicsAddVelocity(builder, velocity): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(velocity), 0)
def PhysicsAddAngularVelocity(builder, angularVelocity): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(angularVelocity), 0)
def PhysicsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class Touch(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTouch(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Touch()
        x.Init(buf, n + offset)
        return x

    # Touch
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// The name of the player involved with the touch.
    # Touch
    def PlayerName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

# /// Seconds that had elapsed in the game when the touch occurred.
    # Touch
    def GameSeconds(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// The point of contact for the touch.
    # Touch
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// The direction of the touch.
    # Touch
    def Normal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// The Team which the touch belongs to, 0 for blue 1 for orange.
    # Touch
    def Team(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

# /// The index of the player involved with the touch.
    # Touch
    def PlayerIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def TouchStart(builder): builder.StartObject(6)
def TouchAddPlayerName(builder, playerName): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(playerName), 0)
def TouchAddGameSeconds(builder, gameSeconds): builder.PrependFloat32Slot(1, gameSeconds, 0.0)
def TouchAddLocation(builder, location): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def TouchAddNormal(builder, normal): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(normal), 0)
def TouchAddTeam(builder, team): builder.PrependInt32Slot(4, team, 0)
def TouchAddPlayerIndex(builder, playerIndex): builder.PrependInt32Slot(5, playerIndex, 0)
def TouchEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// A minimal version of player data, useful when bandwidth needs to be conserved.
class TinyPlayer(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTinyPlayer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TinyPlayer()
        x.Init(buf, n + offset)
        return x

    # TinyPlayer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TinyPlayer
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TinyPlayer
    def Rotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Rotator import Rotator
            obj = Rotator()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TinyPlayer
    def Velocity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TinyPlayer
    def HasWheelContact(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # TinyPlayer
    def IsSupersonic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # TinyPlayer
    def Team(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # TinyPlayer
    def Boost(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def TinyPlayerStart(builder): builder.StartObject(7)
def TinyPlayerAddLocation(builder, location): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def TinyPlayerAddRotation(builder, rotation): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(rotation), 0)
def TinyPlayerAddVelocity(builder, velocity): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(velocity), 0)
def TinyPlayerAddHasWheelContact(builder, hasWheelContact): builder.PrependBoolSlot(3, hasWheelContact, 0)
def TinyPlayerAddIsSupersonic(builder, isSupersonic): builder.PrependBoolSlot(4, isSupersonic, 0)
def TinyPlayerAddTeam(builder, team): builder.PrependInt32Slot(5, team, 0)
def TinyPlayerAddBoost(builder, boost): builder.PrependInt32Slot(6, boost, 0)
def TinyPlayerEnd(builder): return builder.EndObject()



import multiprocessing
import os
import socket
import subprocess
import sys
from typing import Type

import psutil
from rlbot.agents.base_agent import BOT_CONFIG_AGENT_HEADER
from rlbot.agents.executable_with_socket_agent import ExecutableWithSocketAgent
from rlbot.agents.standalone.standalone_bot_config import StandaloneBotConfig
from rlbot.parsing.bot_config_bundle import get_bot_config_bundle
from rlbot.utils.process_configuration import configure_process, get_team_cpus


class ExecutableWithSocketStandaloneBot(ExecutableWithSocketAgent):

    def __init__(self, name, team, index):
        super().__init__(name, team, index)

    def find_process_using_port(self, port: int):
        for proc in psutil.process_iter():
            for conn in proc.connections():
                if conn.laddr.port == port:
                    return proc
        return None

    def is_port_open(self, port: int):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect(("127.0.0.1", port))
            s.close()
            return True
        except ConnectionRefusedError:
            s.close()
            return False

    def send_add_command(self):
        message = self.build_add_command()
        self.logger.info(f"About to send add message for {self.name}")
        return self.send_command(message)

    def launch_executable(self):
        self.logger.info("Launching executable...")
        launch = [self.executable_path, str(self.get_port())]
        proc = subprocess.Popen(launch, cwd=os.path.dirname(self.executable_path))
        process = psutil.Process(pid=proc.pid)
        configure_process(process, get_team_cpus(self.team), infer_multi_team=True)
        self.logger.info("Launched executable instance")


def run_bot(agent_class: Type[ExecutableWithSocketStandaloneBot]):
    config = StandaloneBotConfig(sys.argv)

    bundle = get_bot_config_bundle(config.config_file)
    config_obj = agent_class.base_create_agent_configurations()
    config_obj.parse_file(bundle.config_obj, config_directory=bundle.config_directory)

    agent = agent_class(config.name, config.team, config.player_index)
    agent.spawn_id = config.spawn_id
    agent.load_config(config_obj.get_header(BOT_CONFIG_AGENT_HEADER))
    if not agent.send_add_command():
        agent.launch_executable()
    agent.run_independently(multiprocessing.Event())
    agent.retire()



"""Standalone Bot Script

Usage:
    standalone [--player-index=0] [--config-file=C:/Users/t/code/myBot.cfg] [--name=MyBot] [--team=0] [--spawn-id=23245] [--matchcomms-url=ws://localhost:49468]

Options:
    --player-index=I             Index that the player is running under. Will be passed to your bot's constructor.
    --config-file=C              Absolute path to .cfg file, which could define the executable path, preferred tick rate, etc.
    --name=N                     Name that will be passed to your bot's constructor. Does not influence in-game name.
    --team=T                     Team the bot is playing on, 0 for blue, 1 for orange. Will be passed to your bot's constructor.
    --spawn-id=S                 Spawn identifier used to confirm the right car in the packet.
    --matchcomms-url=M           A url that can be used to connect with the matchcomms system.
"""
from typing import List, Union

from docopt import docopt, DocoptExit, printable_usage
from urllib.parse import ParseResult, urlparse


class StandaloneBotConfig:

    def __init__(self, argv: List[str], strict_mode=True):
        arguments = docopt(__doc__, argv[1:])
        self.name = arguments['--name']
        self.team = self.int_or_none(arguments['--team'])
        self.player_index = self.int_or_none(arguments['--player-index'])
        self.spawn_id = self.int_or_none(arguments['--spawn-id'])
        self.config_file = arguments['--config-file']
        self.matchcomms_url = self.url_or_none(arguments['--matchcomms-url'])
        self.is_missing_args = False

        if self.name is None or self.team is None or self.player_index is None or self.spawn_id is None \
                or self.config_file is None or self.matchcomms_url is None:
            print('Standalone bot is missing required arguments!')
            print(printable_usage(__doc__))
            self.is_missing_args = True

    def int_or_none(self, val) -> Union[int, None]:
        if val:
            return int(val)
        return None

    def url_or_none(self, val) -> Union[ParseResult, None]:
        if val:
            return urlparse(val)
        return None



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class PlayerInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerInfo()
        x.Init(buf, n + offset)
        return x

    # PlayerInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PlayerInfo
    def Physics(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Physics import Physics
            obj = Physics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PlayerInfo
    def ScoreInfo(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .ScoreInfo import ScoreInfo
            obj = ScoreInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PlayerInfo
    def IsDemolished(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// True if your wheels are on the ground, the wall, or the ceiling. False if you're midair or turtling.
    # PlayerInfo
    def HasWheelContact(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # PlayerInfo
    def IsSupersonic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # PlayerInfo
    def IsBot(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// True if the player has jumped. Falling off the ceiling / driving off the goal post does not count.
    # PlayerInfo
    def Jumped(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# ///  True if player has double jumped. False does not mean you have a jump remaining, because the
# ///  aerial timer can run out, and that doesn't affect this flag.
    # PlayerInfo
    def DoubleJumped(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # PlayerInfo
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

    # PlayerInfo
    def Team(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerInfo
    def Boost(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerInfo
    def Hitbox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .BoxShape import BoxShape
            obj = BoxShape()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PlayerInfo
    def HitboxOffset(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

# /// In the case where the requested player index is not available, spawnId will help
# /// the framework figure out what index was actually assigned to this player instead.
    # PlayerInfo
    def SpawnId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def PlayerInfoStart(builder): builder.StartObject(14)
def PlayerInfoAddPhysics(builder, physics): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(physics), 0)
def PlayerInfoAddScoreInfo(builder, scoreInfo): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(scoreInfo), 0)
def PlayerInfoAddIsDemolished(builder, isDemolished): builder.PrependBoolSlot(2, isDemolished, 0)
def PlayerInfoAddHasWheelContact(builder, hasWheelContact): builder.PrependBoolSlot(3, hasWheelContact, 0)
def PlayerInfoAddIsSupersonic(builder, isSupersonic): builder.PrependBoolSlot(4, isSupersonic, 0)
def PlayerInfoAddIsBot(builder, isBot): builder.PrependBoolSlot(5, isBot, 0)
def PlayerInfoAddJumped(builder, jumped): builder.PrependBoolSlot(6, jumped, 0)
def PlayerInfoAddDoubleJumped(builder, doubleJumped): builder.PrependBoolSlot(7, doubleJumped, 0)
def PlayerInfoAddName(builder, name): builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def PlayerInfoAddTeam(builder, team): builder.PrependInt32Slot(9, team, 0)
def PlayerInfoAddBoost(builder, boost): builder.PrependInt32Slot(10, boost, 0)
def PlayerInfoAddHitbox(builder, hitbox): builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(hitbox), 0)
def PlayerInfoAddHitboxOffset(builder, hitboxOffset): builder.PrependStructSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(hitboxOffset), 0)
def PlayerInfoAddSpawnId(builder, spawnId): builder.PrependInt32Slot(13, spawnId, 0)
def PlayerInfoEnd(builder): return builder.EndObject()



from time import sleep

from rlbot.agents.base_agent import BaseAgent
from rlbot.botmanager.bot_manager import BotManager
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.structures import ball_prediction_struct as bp
from rlbot.utils.structures import game_data_struct as gd
from rlbot.utils.structures.bot_input_struct import PlayerInput
from rlbot.utils.structures.rigid_body_struct import RigidBodyTick


class BotManagerStruct(BotManager):
    def __init__(self, *args, **kwargs):
        """
        See documentation on BotManager.
        """
        super().__init__(*args, **kwargs)
        self.rigid_body_tick = None

    def prepare_for_run(self):
        # Set up shared memory map (offset makes it so bot only writes to its own input!) and map to buffer
        self.bot_input = PlayerInput()
        # Set up shared memory for game data
        # We want to do a deep copy for game inputs so people don't mess with em
        self.game_tick_packet = gd.GameTickPacket()
        # Set up shared memory for Ball prediction
        self.ball_prediction = bp.BallPrediction()
        # Set up shared memory for rigid body tick
        self.rigid_body_tick = RigidBodyTick()

    def call_agent(self, agent: BaseAgent, agent_class):
        # Call begin rendering on the agent's behalf. If they call begin_rendering again internally, no harm done.
        agent.renderer.begin_rendering()

        controller_input = agent.get_output(self.game_tick_packet)
        if controller_input is None:
            get_logger("BotManager" + str(self.index))\
                .error("Agent %s did not return any output.", str(agent_class.__name__))
            return

        player_input = self.bot_input

        if isinstance(controller_input, list):
            # Write all player inputs
            get_logger("BotManager" + str(self.index)).error("Sending legacy packet type, please convert to v4")
            controller_input = agent.convert_output_to_v4(controller_input)

        player_input.throttle = controller_input.throttle
        player_input.steer = controller_input.steer
        player_input.pitch = controller_input.pitch
        player_input.yaw = controller_input.yaw
        player_input.roll = controller_input.roll
        player_input.jump = controller_input.jump
        player_input.boost = controller_input.boost
        player_input.handbrake = controller_input.handbrake
        if hasattr(controller_input, 'use_item'):
            # This is needed for rare cases where bots don't conform to the spec,
            # e.g. Stick returns itself rather than a SimpleControllerState.
            player_input.use_item = controller_input.use_item
        self.game_interface.update_player_input(player_input, self.index)

        # Send the rendering at the end, to minimize any impact on latency. The agent may have already called
        # end_rendering internally, if that's the case then this will just no-op.
        agent.renderer.end_rendering()

    def get_game_time(self):
        return self.game_tick_packet.game_info.seconds_elapsed

    def pull_data_from_game(self):
        if self.match_config.enable_lockstep:
            # If we're using lockstep then fresh_live_data_packet is no good because when
            # lockstep freezes the game, we won't get a fresh packet and it'll wait the full timeout.
            # Instead do a clunky sleep here to limit the framerate. Not worried about precision because
            # lockstep will probably never be used in a competitive scenario.
            sleep(1 / self.maximum_tick_rate_preference)
            self.game_interface.update_live_data_packet(self.game_tick_packet)
        else:
            # Set a timeout of 30 milliseconds. It's slightly less than the number of milliseconds (33.33)
            # caused by MAX_AGENT_CALL_PERIOD defined in bot_manager.py
            self.game_interface.fresh_live_data_packet(self.game_tick_packet, 30, self.index)

    def get_ball_prediction_struct(self):
        return self.game_interface.update_ball_prediction(self.ball_prediction)

    def get_spawn_id(self):
        if self.game_tick_packet.num_cars <= self.index:
            return None
        return self.game_tick_packet.game_cars[self.index].spawn_id



import flatbuffers
import multiprocessing
import queue
from threading import Thread

from rlbot.messages.flat import QuickChat
from rlbot.messages.flat import QuickChatSelection
from rlbot.utils.logging_utils import get_logger

from rlbot.utils.structures.utils import create_enum_object


"""
Look for quick chats from here:
https://github.com/RLBot/RLBot/blob/master/src/main/flatbuffers/rlbot.fbs
"""
QuickChats = create_enum_object([chat for chat in dir(QuickChatSelection.QuickChatSelection)
                                 if not chat.startswith('__') and not
                                 callable(getattr(QuickChatSelection.QuickChatSelection, chat))],
                                list_name='quick_chat_list',
                                other_attributes=[
                                    ('CHAT_NONE', -1),
                                    ('CHAT_EVERYONE', False),
                                    ('CHAT_TEAM_ONLY', True)
],
    attribute_object=QuickChatSelection.QuickChatSelection)


def send_quick_chat_flat(game_interface, index, team, team_only, quick_chat):
    builder = flatbuffers.Builder(0)
    QuickChat.QuickChatStart(builder)
    QuickChat.QuickChatAddQuickChatSelection(builder, quick_chat)
    QuickChat.QuickChatAddPlayerIndex(builder, index)
    QuickChat.QuickChatAddTeamOnly(builder, team_only)
    result = QuickChat.QuickChatEnd(builder)

    builder.Finish(result)

    return game_interface.send_chat_flat(builder)


def send_quick_chat(queue_holder, index, team, team_only, quick_chat):
    """
    Sends a quick chat to the general queue for everyone to pull from
    :param queue_holder:
    :param index: The index of the player sending the message
    :param team: The team of the player sending the message
    :param team_only: if the message is team only
    :param quick_chat: The contents of the quick chat
    :return:
    """
    queue_holder["output"].put((index, team, team_only, quick_chat))
